import numpy as np
import scipy.ndimage as ndimage
import matplotlib.pyplot as plt
from matplotlib import patches
import matplotlib
from shapely import geometry
import PIL
from PIL import ImageDraw
import math
import scipy.stats as stats
# from matplotlib.colors import NoNorm
import matplotlib.cm as cm
import networkx as nx
from deepcinac.utils.utils import welsh_powell
from deepcinac.utils.display import BREWER_COLORS
import matplotlib.colors as plt_colors
from shapely.geometry import MultiPoint, LineString
import os


def create_cells_coord_from_suite_2p(is_cell_file_name, stat_file_name, movie_dimensions):
    """

    Args:
        is_cell_file_name: path and file_name of the file iscell.npy produce by suite2p segmentation process
        Could be None, if None, means all cells in stat are cells to keep
        stat_file_name: path and file_name of the file stat.npy produce by suite2p segmentation process
        movie_dimensions: tuple of 2 int, dimensions of the movie n col x n line

    Returns: a CellsCoord instance

    """
    stat = np.load(os.path.join(stat_file_name), allow_pickle=True)

    if is_cell_file_name is None:
        is_cell = None
    else:
        is_cell = np.load(os.path.join(is_cell_file_name), allow_pickle=True)

    coords = []
    for cell in np.arange(len(stat)):
        if is_cell is not None:
            if is_cell[cell][0] == 0:
                continue
        list_points_coord = [(x, y) for x, y in zip(stat[cell]["xpix"], stat[cell]["ypix"])]
        # TODO: pass the pixel_mask shape directly when available in CellsCoord
        convex_hull = MultiPoint(list_points_coord).convex_hull
        if isinstance(convex_hull, LineString):
            coord_shapely = MultiPoint(list_points_coord).convex_hull.coords
        else:
            coord_shapely = MultiPoint(list_points_coord).convex_hull.exterior.coords
        coords.append(np.array(coord_shapely).transpose())

    return CellsCoord(coords=coords, nb_lines=movie_dimensions[0], nb_col=movie_dimensions[1], from_matlab=False)


class CellsCoord:

    def __init__(self, coords=None, pixel_masks=None, nb_lines=None, nb_col=None,
                 from_matlab=False, invert_xy_coord=False):  # from_suite_2p=False, from_fiji=False
        """

        Args:
            coords: list of array of 2 lines (x & y) and n columns, n being the number of coordinate to set the contour of
            the cell. if None, then pixel_masks must be not None
            pixel_masks: a list of list of tuples of int representing the coordinates of the pixels in a given cell.
                len(pixel_masks) == number of cells.
            nb_lines: If None, set it using the max coordinate in coords
            nb_col:If None, set it using the max coordinate in coords
            from_matlab: Means coordinates have been computed on matlab and indexing starts at 1 so
            invert_xy_coord: if True, invert the xy coords
        """

        if coords is None and pixel_masks is None:
            raise Exception("Coords or pixel_masks must be given to instanciate CellsCoord")

        # dict of tuples, key is the cell #, cell center coords x and y (x and y are inverted for imshow)
        self.center_coord = dict()
        # shapely polygons
        self.cells_polygon = dict()
        # first key is an int representing the number of the cell, and value is a list of cells it interesects
        self.intersect_cells = dict()
        # for garbage collector issue only
        self.cell_contour = None

        # contour coords
        self.coords = coords

        # ---------------------------
        # we clean coords if not None
        # ---------------------------
        if self.coords is not None:
            for cell, coord in enumerate(self.coords):
                # coord = np.array(coord)
                # print(f"coord {coord}")
                if from_matlab:
                    # it is necessary to remove one, as data comes from matlab, starting from 1 and not 0
                    coord = coord - 1
                # in case it would be floats
                coord.astype(int, copy=False)
                if invert_xy_coord:
                    # invert xy lines
                    coord = np.flipud(coord)
                self.coords[cell] = coord

        self.pixel_masks = pixel_masks

        # ---------------------------
        # we clean pixel_masks if not None
        # we change the structure of so each element
        # is an np.array of 2d with 2 lines and n columns
        # n being the number of pixels
        # ---------------------------
        if self.pixel_masks is not None:
            tmp_pixel_masks = self.pixel_masks
            self.pixel_masks = []
            to_substract = 0
            if from_matlab:
                to_substract = 1
            for pixel_mask in tmp_pixel_masks:
                pix_array = np.zeros((2, len(pixel_mask)), dtype="int16")
                for index_pix, pix in enumerate(pixel_mask):
                    if invert_xy_coord:
                        pix_array[0, index_pix] = int(pix[1] - to_substract)
                        pix_array[1, index_pix] = int(pix[0] - to_substract)
                    else:
                        pix_array[0, index_pix] = int(pix[0] - to_substract)
                        pix_array[1, index_pix] = int(pix[1] - to_substract)
                self.pixel_masks.append(pix_array)

        # --------------------------------------------
        # Number of lines and columsn (height & width)
        # of the Field Of View (FOV)
        # --------------------------------------------
        if (nb_lines is None) or (nb_col is None):
            # in that case we take the maximum coordinates to fix the size of the FOV adding some padding
            self.nb_lines = 0
            self.nb_col = 0
            padding = 2

            if self.coords is not None:
                for cell, coord in enumerate(self.coords):
                    self.nb_col = max(self.nb_col, np.max(coord[0, :]) + padding)
                    self.nb_lines = max(self.nb_lines, np.max(coord[1, :]) + padding)
            else:
                for pixel_mask in self.pixel_masks:
                    # raise Exception("TOTO")
                    self.nb_col = max(self.nb_col, np.max(pixel_mask[0, :]) + padding)
                    self.nb_lines = max(self.nb_lines, np.max(pixel_mask[1, :]) + padding)
        else:
            self.nb_lines = nb_lines
            self.nb_col = nb_col

        if self.coords is not None:
            self.n_cells = len(self.coords)
        else:
            self.n_cells = len(self.pixel_masks)

        # -----------------------------------------------------------
        # Using pixels_masks to build contour coords if doesn't exist
        # -----------------------------------------------------------
        if self.pixel_masks is not None:
            if self.coords is None:
                # first we build self.coords from self.pixel_masks
                self.coords = []
                for cell, pixel_mask in enumerate(self.pixel_masks):
                    # we use pixel_masks to build the coords, using convex_hull
                    list_points_coord = list(zip(pixel_mask[1], pixel_mask[0]))
                    convex_hull = MultiPoint(list_points_coord).convex_hull
                    if isinstance(convex_hull, LineString):
                        coord_shapely = MultiPoint(list_points_coord).convex_hull.coords
                    else:
                        coord_shapely = MultiPoint(list_points_coord).convex_hull.exterior.coords
                    self.coords.append(np.array(coord_shapely).transpose())

        # ----------------------------------------------------------------------
        # Going through contour coords and building polygons and centers of mass
        # ----------------------------------------------------------------------
        # print(f"self.coords {self.coords}")
        for cell, coord in enumerate(self.coords):
            if coord.shape[0] == 0:
                print(f'Error: {cell} coord.shape {coord.shape}')
                continue

            self.build_cell_polygon_from_contour(cell=cell)

            # poly = MultiPoint(list_points).convex_hull

            use_centroid = True

            if use_centroid:
                self.center_coord[cell] = self.cells_polygon[cell].centroid.coords[0]
            else:
                bw = np.zeros((self.nb_col, self.nb_lines), dtype="int8")
                # we could use morphology.binary_fill_holes to build pixel_mask from coord
                # but let's keep pixel_mask to None if not passed as argument
                # morphology.binary_fill_holes(input
                bw[coord[0, :], coord[1, :]] = 1
                c_x, c_y = ndimage.center_of_mass(bw)
                self.center_coord[cell] = (c_x, c_y)

            # if (cell == 0) or (cell == 159):
            #     print(f"cell {cell} fig")
            #     fig, ax = plt.subplots(nrows=1, ncols=1,
            #                            gridspec_kw={'height_ratios': [1]},
            #                            figsize=(5, 5))
            #     ax.imshow(bw)
            #     plt.show()
            #     plt.close()

        # --------------------------------------------------------------
        # building the intersections dictionnary using coords
        # --------------------------------------------------------------
        for cell_1 in np.arange(self.n_cells - 1):
            if cell_1 not in self.intersect_cells:
                if cell_1 not in self.cells_polygon:
                    continue
                self.intersect_cells[cell_1] = set()
            for cell_2 in np.arange(cell_1 + 1, self.n_cells):
                if cell_2 not in self.cells_polygon:
                    continue
                if cell_2 not in self.intersect_cells:
                    self.intersect_cells[cell_2] = set()
                poly_1 = self.cells_polygon[cell_1]
                poly_2 = self.cells_polygon[cell_2]
                # if it intersects and not only touches if adding and (not poly_1.touches(poly_2))
                # try:
                if poly_1.intersects(poly_2):
                    self.intersect_cells[cell_2].add(cell_1)
                    self.intersect_cells[cell_1].add(cell_2)
                # except shapely.errors.TopologicalError:
                #     print(f"cell_1 {cell_1}, cell_2 {cell_2}")
                #     print(f"cell_1 {poly_1.is_valid}, cell_2 {poly_2.is_valid}")
                #     poly_1 = poly_1.buffer(0)
                #     poly_2 = poly_2.buffer(0)
                #     print(f"cell_1 {poly_1.is_valid}, cell_2 {poly_2.is_valid}")
                #     raise Exception("shapely.errors.TopologicalError")

    def build_raw_traces_from_movie(self, movie):
        """
        Return a 2d array representing the fluoresence signal raw trace for each cell
        Args:
            movie: 3d array n_frames x len_y x len_x

        Returns: A 2d array (n_cells * n_frames) of float

        """
        raw_traces = np.zeros((self.n_cells, movie.shape[0]))
        for cell in np.arange(self.n_cells):
            mask = self.get_cell_mask(cell=cell,
                                      dimensions=(movie.shape[1], movie.shape[2]))
            raw_traces[cell, :] = np.mean(movie[:, mask], axis=1)
        return raw_traces

    def build_cell_polygon_from_contour(self, cell):
        """
        Build the (shapely) polygon representing a given cell using its contour's coordinates.
        Args:
            cell:

        Returns:

        """
        coord = self.coords[cell]
        # make a list of tuple representing x, y coordinates of the contours points
        coord_list_tuple = list(zip(coord[0], coord[1]))

        # buffer(0) or convex_hull could be used if the coords are a list of points not
        # in the right order. However buffer(0) return a MultiPolygon with no coords available.
        if len(coord_list_tuple) < 3:
            list_points = []
            for coords in coord_list_tuple:
                list_points.append(geometry.Point(coords))
            self.cells_polygon[cell] = geometry.LineString(list_points)
        else:
            self.cells_polygon[cell] = geometry.Polygon(coord_list_tuple)  # .convex_hull # buffer(0)

        # self.coords[cell] = np.array(self.cells_polygon[cell].exterior.coords).transpose()

    def get_cell_mask(self, cell, dimensions):
        """
        :param cell:
        :param dimensions: height x width
        :return:
        """
        poly_gon = self.cells_polygon[cell]
        img = PIL.Image.new('1', (dimensions[1], dimensions[0]), 0)
        if isinstance(poly_gon, geometry.LineString):
            ImageDraw.Draw(img).polygon(list(poly_gon.coords), outline=1,
                                        fill=1)
        else:
            ImageDraw.Draw(img).polygon(list(poly_gon.exterior.coords), outline=1,
                                        fill=1)
        return np.array(img)

    def match_cells_indices(self, coord_obj, path_results, plot_title_opt=""):
        """

        :param coord_obj: another instanc of coord_obj
        :return: a 1d array, each index corresponds to the index of a cell of coord_obj, and map it to an index to self
        or -1 if no cell match
        """
        mapping_array = np.zeros(len(coord_obj.coord), dtype='int16')
        for cell in np.arange(len(coord_obj.coord)):
            c_x, c_y = coord_obj.center_coord[cell]
            distances = np.zeros(len(self.coords))
            for self_cell in np.arange(len(self.coords)):
                self_c_x, self_c_y = self.center_coord[self_cell]
                # then we calculte the cartesian distance to all other cells
                distances[self_cell] = math.sqrt((self_c_x - c_x) ** 2 + (self_c_y - c_y) ** 2)
            if np.min(distances) <= 2:
                mapping_array[cell] = np.argmin(distances)
            else:
                mapping_array[cell] = -1
        plot_result = True
        if plot_result:
            fig, ax = plt.subplots(nrows=1, ncols=1,
                                   gridspec_kw={'height_ratios': [1]},
                                   figsize=(20, 20))

            fig.patch.set_facecolor("black")
            ax.set_facecolor("black")

            # dark blue
            other_twin_color = list((0.003, 0.313, 0.678, 1.0))
            n_twins = 0
            # red
            other_orphan_color = list((1, 0, 0, 1.0))
            n_other_orphans = 0
            # light blue
            self_twin_color = list((0.560, 0.764, 1, 1.0))
            # green
            self_orphan_color = list((0.278, 1, 0.101, 1.0))
            n_self_orphans = 0
            with_edge = True
            edge_line_width = 1
            z_order_cells = 12
            for cell in np.arange(len(coord_obj.coord)):
                xy = coord_obj.coord[cell].transpose()
                if with_edge:
                    line_width = edge_line_width
                    edge_color = "white"
                else:
                    edge_color = "white"
                    line_width = 0
                # allow to set alpha of the edge to 1
                if mapping_array[cell] >= 0:
                    # dark blue
                    face_color = other_twin_color
                    n_twins += 1
                else:
                    # red
                    face_color = other_orphan_color
                    n_other_orphans += 1
                face_color[3] = 0.8
                face_color = tuple(face_color)
                cell_contour = patches.Polygon(xy=xy,
                                               fill=True, linewidth=line_width,
                                               facecolor=face_color,
                                               edgecolor=edge_color,
                                               zorder=z_order_cells)  # lw=2
                ax.add_patch(cell_contour)
            for cell in np.arange(len(self.coords)):
                xy = self.coords[cell].transpose()
                if with_edge:
                    line_width = edge_line_width
                    edge_color = "white"
                else:
                    edge_color = "white"
                    line_width = 0
                # allow to set alpha of the edge to 1
                if cell in mapping_array:
                    # light blue
                    face_color = self_twin_color
                else:
                    # green
                    face_color = self_orphan_color
                    n_self_orphans += 1
                face_color[3] = 0.8
                face_color = tuple(face_color)
                # drawing the cell
                cell_contour = patches.Polygon(xy=xy,
                                               fill=True, linewidth=line_width,
                                               facecolor=face_color,
                                               edgecolor=edge_color,
                                               zorder=z_order_cells)  # lw=2
                ax.add_patch(cell_contour)
                with_cell_numbers = True
                if with_cell_numbers:
                    self.plot_text_cell(cell=cell, cell_numbers_color="black", ax=ax, text_size=6)
            fontsize = 12
            plt.text(x=190, y=180,
                     s=f"{n_twins}", color=self_twin_color, zorder=22,
                     ha='center', va="center", fontsize=fontsize, fontweight='bold')
            plt.text(x=190, y=185,
                     s=f"{n_self_orphans}", color=self_orphan_color, zorder=22,
                     ha='center', va="center", fontsize=fontsize, fontweight='bold')
            plt.text(x=190, y=190,
                     s=f"{n_twins}", color=other_twin_color, zorder=22,
                     ha='center', va="center", fontsize=fontsize, fontweight='bold')
            plt.text(x=190, y=195,
                     s=f"{n_other_orphans}", color=other_orphan_color, zorder=22,
                     ha='center', va="center", fontsize=fontsize, fontweight='bold')
            ax.set_ylim(0, self.nb_lines)
            ax.set_xlim(0, self.nb_col)
            ylim = ax.get_ylim()
            # invert Y
            ax.set_ylim(ylim[::-1])
            plt.setp(ax.spines.values(), color="black")
            frame = plt.gca()
            frame.axes.get_xaxis().set_visible(False)
            frame.axes.get_yaxis().set_visible(False)
            save_format = "png"
            fig.savefig(f'{path_results}/cells_map_{plot_title_opt}.{save_format}',
                        format=f"{save_format}",
                        facecolor=fig.get_facecolor())
            plt.close()
        return mapping_array

    def plot_cells_map(self, path_results, data_id, use_pixel_masks=False, title_option="", connections_dict=None,
                       background_color=(0, 0, 0, 1), default_cells_color=(1, 1, 1, 1.0),
                       default_edge_color="white",
                       dont_fill_cells_not_in_groups=False,
                       link_connect_color="white", link_line_width=1,
                       cell_numbers_color="dimgray", show_polygons=False,
                       cells_to_link=None, edge_line_width=2, cells_alpha=1.0,
                       fill_polygons=True, cells_groups=None, cells_groups_colors=None,
                       cells_groups_alpha=None,
                       cells_to_hide=None, img_on_background=None,
                       real_size_image_on_bg=True,
                       cells_groups_edge_colors=None, with_edge=False,
                       with_cell_numbers=False, text_size=6, save_formats="png",
                       save_plot=True, return_fig=False, ax_to_use=None,
                       verbose=False,
                       use_welsh_powell_coloring=False, dpi=300):
        """

        Args:
            path_results:
            data_id:
            use_pixel_masks:
            title_option:
            connections_dict: key is an int representing a cell number, and value is a dict representing the cells it
            connects to. The key is a cell is connected too, and the value represent the strength of the connection
            (like how many times it connects to it)
            background_color:
            default_cells_color:
            default_edge_color:
            dont_fill_cells_not_in_groups:
            link_connect_color:
            link_line_width:
            cell_numbers_color:
            show_polygons:
            cells_to_link:
            edge_line_width:
            cells_alpha:
            fill_polygons:
            cells_groups:
            cells_groups_colors:
            cells_groups_alpha:
            cells_to_hide:
            img_on_background:
            real_size_image_on_bg: if True, the size of the figure will respect the original size of the background
            image
            cells_groups_edge_colors:
            with_edge:
            with_cell_numbers:
            text_size:
            save_formats:
            save_plot:
            return_fig:
            ax_to_use:
            verbose: if True, some informations will be printed along the way
            use_welsh_powell_coloring: if True, use welsh powell algorithm to color all cells that intersect with
        different color. In that case, cancel cell_groups arguments.
            dpi:

        Returns:

        """

        if use_pixel_masks and (self.pixel_masks is None):
            print(f"No pixel_masks available in plot_cells_map() and use_pixel_masks argument set to True")
            return

        cells_center = self.center_coord
        n_cells = len(self.coords)
        if cells_to_hide is None:
            cells_to_hide = []

        if use_welsh_powell_coloring:
            if verbose:
                print("Welsh Powell coloring:")
            isolated_cell_color = default_cells_color
            isolated_group = []
            cells_groups_colors = []
            cells_groups_edge_colors = []
            cells_groups_alpha = []
            cells_groups = []

            # building networkx graph
            graphs = []
            cells_added = []
            for cell in np.arange(n_cells):
                if cell in cells_added:
                    continue
                # welsh_powell
                n_intersect = len(self.intersect_cells[cell])
                if n_intersect == 0:
                    isolated_group.append(cell)
                    cells_added.append(cell)
                else:
                    graph = nx.Graph()
                    cells_to_expend = [cell]
                    edges = set()
                    while len(cells_to_expend) > 0:
                        if cells_to_expend[0] not in cells_added:
                            cells_added.append(cells_to_expend[0])
                            n_intersect = len(self.intersect_cells[cells_to_expend[0]])
                            if n_intersect > 0:
                                for inter_cell in self.intersect_cells[cells_to_expend[0]]:
                                    min_c = min(inter_cell, cells_to_expend[0])
                                    max_c = max(inter_cell, cells_to_expend[0])
                                    edges.add((min_c, max_c))
                                    cells_to_expend.append(inter_cell)
                        cells_to_expend = cells_to_expend[1:]
                    graph.add_edges_from(list(edges))
                    graphs.append(graph)
            cells_by_color_code = dict()
            max_color_code = 0
            for graph in graphs:
                # dict that give for each cell a color code
                col_val = welsh_powell(graph)
                for cell, color_code in col_val.items():
                    if color_code not in cells_by_color_code:
                        cells_by_color_code[color_code] = []
                    cells_by_color_code[color_code].append(cell)
                    max_color_code = max(max_color_code, color_code)

            for color_code, cells in cells_by_color_code.items():
                if len(cells) == 0:
                    continue
                cells_groups.append(cells)
                # cells_groups_colors.append(cm.nipy_spectral(float(color_code + 1) / (max_color_code + 1)))
                color_hex = BREWER_COLORS[color_code % len(BREWER_COLORS)]
                # putting it with value from 0.0 to 1.0
                color_rgb = plt_colors.hex2color(color_hex)
                # adding the alpha
                color_rgba = [c for c in color_rgb]
                color_rgba.append(cells_alpha)
                cells_groups_colors.append(color_rgba)
                cells_groups_edge_colors.append(default_edge_color)
                cells_groups_alpha.append(0.8)
            cells_groups.append(isolated_group)
            cells_groups_colors.append(isolated_cell_color)
            cells_groups_alpha.append(1)
            cells_groups_edge_colors.append("white")
            if verbose:
                print(f"Isolated cells: {len(isolated_group)}")
                print(f"Grouped cells: {self.n_cells - len(isolated_group)}")

        cells_in_groups = []
        if cells_groups is not None:
            for group_id, cells_group in enumerate(cells_groups):
                cells_in_groups.extend(cells_group)
        cells_in_groups = np.array(cells_in_groups)
        cells_not_in_groups = np.setdiff1d(np.arange(n_cells), cells_in_groups)

        if ax_to_use is None:
            if (img_on_background is None) or (real_size_image_on_bg is False):
                fig, ax = plt.subplots(nrows=1, ncols=1,
                                       gridspec_kw={'height_ratios': [1]},
                                       figsize=(20, 20), dpi=dpi)
                fig.subplots_adjust(bottom=0, top=1, left=0, right=1)
                plt.tight_layout()
            else:
                # then we want the figure to respect the size of the image on background
                height, width = img_on_background.shape

                # What size does the figure need to be in inches to fit the image?
                figsize = width / float(dpi), height / float(dpi)

                # Create a figure of the right size with one axes that takes up the full figure
                fig = plt.figure(figsize=figsize)
                ax = fig.add_axes([0, 0, 1, 1])

                text_size = 0.05
                edge_line_width = 0.1

            fig.patch.set_facecolor(background_color)
            ax.set_facecolor(background_color)
        else:
            ax = ax_to_use

        if img_on_background is not None:
            img_on_background = img_on_background - np.min(img_on_background)
            # print(f"np.mean(img_on_background) {np.mean(img_on_background)} {np.max(img_on_background)}")
            n_bits = 8
            # np.transpose(img_on_background)
            ax.imshow(img_on_background, cmap=plt.get_cmap("gray"))  # , vmin=0,
            # vmax=np.max(img_on_background), interpolation='nearest') # vmax= 4096 math.pow(2, n_bits)-1

        if use_pixel_masks:
            cells_imshow_alpha = 0.2 if (img_on_background is not None) else 1
            bg_imshow_alpha = 0 if (img_on_background is not None) else 1

            self.add_cells_using_pixel_masks_on_ax(ax, cells_groups, cells_not_in_groups, cells_to_hide,
                                                   default_cells_color, cells_groups_colors, with_cell_numbers,
                                                   cell_numbers_color, text_size,
                                                   background_color, cells_imshow_alpha=cells_imshow_alpha,
                                                   bg_imshow_alpha=bg_imshow_alpha)
        else:
            self.add_cells_using_polygons_on_ax(ax, cells_groups, cells_not_in_groups, cells_to_hide, with_edge,
                                                edge_line_width, default_cells_color,
                                                default_edge_color, cells_groups_edge_colors, cells_groups_colors,
                                                cells_groups_alpha, cells_alpha, with_cell_numbers,
                                                cell_numbers_color, text_size, dont_fill_cells_not_in_groups)

        ax.set_ylim(0, self.nb_lines)
        ax.set_xlim(0, self.nb_col)
        ylim = ax.get_ylim()
        # invert Y
        ax.set_ylim(ylim[::-1])

        if connections_dict is not None:
            zorder_lines = 15
            for neuron in connections_dict.keys():
                # plot a line to all out of the neuron
                for connected_neuron, nb_connexion in connections_dict[neuron].items():
                    line_width = link_line_width + np.log(nb_connexion)

                    c_x = cells_center[neuron][0]
                    c_y = cells_center[neuron][1]
                    c_x_c = cells_center[connected_neuron][0]
                    c_y_c = cells_center[connected_neuron][1]

                    line = plt.plot((c_x, c_x_c), (c_y, c_y_c), linewidth=line_width, c=link_connect_color,
                                    zorder=zorder_lines)[0]
        # print(f"(self.cells_groups is not None) {(self.cells_groups is not None)} show_polygons {show_polygons}")
        if (cells_groups is not None) and show_polygons:
            for group_id, cells in enumerate(cells_groups):
                points = np.zeros((2, len(cells)))
                for cell_id, cell in enumerate(cells):
                    c_x, c_y = cells_center[cell]
                    points[0, cell_id] = c_x
                    points[1, cell_id] = c_y
                # finding the convex_hull for each group
                xy = convex_hull(points=points)
                # xy = xy.transpose()
                # print(f"xy {xy}")
                # xy is a numpy array with as many line as polygon point
                # and 2 columns: x and y coords of each point
                face_color = list(cells_groups_colors[group_id])
                # changing alpha
                face_color[3] = 0.3
                face_color = tuple(face_color)
                # edge alpha will be 1
                poly_gon = patches.Polygon(xy=xy,
                                           fill=fill_polygons, linewidth=0, facecolor=face_color,
                                           edgecolor=cells_groups_colors[group_id],
                                           zorder=15, lw=3)
                ax.add_patch(poly_gon)

        # plt.title(f"Cells map {data_id} {title_option}")

        # ax.set_frame_on(False)

        # invert Y
        ax.set_ylim(ylim[::-1])
        plt.setp(ax.spines.values(), color=background_color)
        frame = plt.gca()
        frame.axes.get_xaxis().set_visible(False)
        frame.axes.get_yaxis().set_visible(False)

        # ax.xaxis.set_ticks_position('none')
        # ax.yaxis.set_ticks_position('none')
        #  :param plot_option: if 0: plot n_out and n_int,
        #  if 1 only n_out, if 2 only n_in, if 3: only n_out with dotted to
        # show the commun n_in and n_out, if 4: only n_in with dotted to show the commun n_in and n_out,
        if ax_to_use is None:
            if save_plot:
                if isinstance(save_formats, str):
                    save_formats = [save_formats]
                for save_format in save_formats:
                    fig.savefig(f'{path_results}/{data_id}_cell_maps_{title_option}.{save_format}',
                                format=f"{save_format}",
                                facecolor=fig.get_facecolor())
            if return_fig:
                return fig
            else:
                plt.close()

    def add_cells_using_pixel_masks_on_ax(self, ax, cells_groups, cells_not_in_groups, cells_to_hide,
                                          default_cells_color, cells_groups_colors, with_cell_numbers,
                                          cell_numbers_color, text_size,
                                          background_color, cells_imshow_alpha, bg_imshow_alpha):
        """
        Using pixel mask if it exists
        Args:
            ax:
            cells_groups:
            cells_not_in_groups:
            cells_to_hide:
            default_cells_color:
            cells_groups_colors:
            with_cell_numbers:
            cell_numbers_color:
            text_size:
            background_color:

        Returns:

        """
        if self.pixel_masks is None:
            return

        use_multiple_imshow = False

        background_color = [code for code in background_color]
        background_color[-1] = bg_imshow_alpha
        cmap_colors = [background_color]
        # 0 represents the background
        color_index = 1
        cells_displayed = []
        if not use_multiple_imshow:
            img = np.zeros((self.nb_col, self.nb_lines), dtype="int8")
        else:
            default_cells_color = [code for code in default_cells_color]
            default_cells_color[-1] = cells_imshow_alpha

        if cells_groups is not None:
            for group_index, cell_group in enumerate(cells_groups):
                for cell in cell_group:
                    if cell in cells_to_hide:
                        continue
                    if use_multiple_imshow:
                        img = np.zeros((self.nb_col, self.nb_lines), dtype="int8")
                        cmap_colors = [background_color]
                        color_index = 1

                    cells_displayed.append(cell)
                    pixel_mask = self.pixel_masks[cell]
                    img[pixel_mask[0, :], pixel_mask[1, :]] = color_index
                    face_color = list(cells_groups_colors[group_index])
                    face_color[-1] = cells_imshow_alpha

                    if use_multiple_imshow:
                        cmap_colors.append(face_color)
                        cmap = matplotlib.colors.ListedColormap(cmap_colors)
                        ax.imshow(np.transpose(img), cmap=cmap, interpolation='nearest')
                if not use_multiple_imshow:
                    cmap_colors.append(face_color)
                    color_index += 1

        for cell in cells_not_in_groups:
            if cell in cells_to_hide:
                continue

            if use_multiple_imshow:
                img = np.zeros((self.nb_col, self.nb_lines), dtype="int8")
                cmap_colors = [background_color]
                color_index = 1

            cells_displayed.append(cell)
            pixel_mask = self.pixel_masks[cell]
            img[pixel_mask[0, :], pixel_mask[1, :]] = color_index

            cmap_colors.append(default_cells_color)
            if use_multiple_imshow:
                cmap = matplotlib.colors.ListedColormap(cmap_colors)
                ax.imshow(np.transpose(img), cmap=cmap, interpolation='nearest')
            else:
                color_index += 1

        if not use_multiple_imshow:
            cmap = matplotlib.colors.ListedColormap(cmap_colors)
            ax.imshow(np.transpose(img), cmap=cmap,
                      interpolation='nearest')  # , origin='lower') alpha=cells_imshow_alpha,

        if with_cell_numbers:
            for cell in cells_displayed:
                self.plot_text_cell(cell=cell, cell_numbers_color=cell_numbers_color, ax=ax,
                                    text_size=text_size)

    def add_cells_using_polygons_on_ax(self, ax, cells_groups, cells_not_in_groups, cells_to_hide, with_edge,
                                       edge_line_width, default_cells_color,
                                       default_edge_color, cells_groups_edge_colors, cells_groups_colors,
                                       cells_groups_alpha, cells_alpha, with_cell_numbers,
                                       cell_numbers_color, text_size, dont_fill_cells_not_in_groups):
        """
        Add cells to a matplolib ax using the polygons representation. Arguments give parameters to apply
        Args:
            ax:
            cells_groups:
            cells_not_in_groups:
            cells_to_hide:
            with_edge:
            edge_line_width:
            default_cells_color:
            default_edge_color:
            cells_groups_edge_colors:
            cells_groups_colors:
            cells_groups_alpha:
            cells_alpha:
            with_cell_numbers:
            cell_numbers_color:
            text_size:
            dont_fill_cells_not_in_groups:

        Returns:

        """
        z_order_cells = 12
        if cells_groups is not None:
            for group_index, cell_group in enumerate(cells_groups):
                for cell in cell_group:
                    if cell in cells_to_hide:
                        continue

                    xy = self.coords[cell].transpose()
                    if with_edge:
                        line_width = edge_line_width
                        if cells_groups_edge_colors is None:
                            edge_color = default_edge_color
                        else:
                            edge_color = cells_groups_edge_colors[group_index]
                    else:
                        edge_color = cells_groups_colors[group_index]
                        line_width = 0
                    # allow to set alpha of the edge to 1
                    face_color = list(cells_groups_colors[group_index])
                    # changing alpha
                    if cells_groups_alpha is not None:
                        face_color[3] = cells_groups_alpha[group_index]
                    else:
                        face_color[3] = cells_alpha
                    face_color = tuple(face_color)
                    # TODO: a solution to explore to plot pixel by pixel and not using the polygon version
                    #  https://stackoverflow.com/questions/39753282/scatter-plot-with-single-pixel-marker-in-matplotlib
                    self.cell_contour = patches.Polygon(xy=xy,
                                                        fill=True, linewidth=line_width,
                                                        facecolor=face_color,
                                                        edgecolor=edge_color,
                                                        zorder=z_order_cells)  # lw=2
                    ax.add_patch(self.cell_contour)
                    if with_cell_numbers:
                        self.plot_text_cell(cell=cell, cell_numbers_color=cell_numbers_color, ax=ax,
                                            text_size=text_size)

        for cell in cells_not_in_groups:
            if cell in cells_to_hide:
                continue
            xy = self.coords[cell].transpose()
            # face_color = default_cells_color
            # if dont_fill_cells_not_in_groups:
            #     face_color = None
            self.cell_contour = patches.Polygon(xy=xy,
                                                fill=not dont_fill_cells_not_in_groups,
                                                linewidth=0, facecolor=default_cells_color,
                                                edgecolor=default_edge_color,
                                                zorder=z_order_cells, lw=edge_line_width)
            ax.add_patch(self.cell_contour)

            if with_cell_numbers:
                self.plot_text_cell(cell=cell, cell_numbers_color=cell_numbers_color, ax=ax,
                                    text_size=text_size)

    def plot_text_cell(self, cell, ax, cell_numbers_color, text_size):
        """
        Plot the cell number on the cell
        Args:
            cell: integer
            ax: matplolib axis
            cell_numbers_color: color of the text
            text_size: text size (float)

        Returns:

        """
        fontsize = text_size
        if cell >= 100:
            if fontsize > 2.5:
                fontsize -= 2
        elif cell >= 10:
            if fontsize > 2:
                fontsize -= 1

        c_x_c = self.center_coord[cell][0]
        c_y_c = self.center_coord[cell][1]

        if fontsize < 0.5:
            fontweight = 'ultralight'
        else:
            fontweight = 'bold'

        ax.text(x=c_x_c, y=c_y_c,
                s=f"{cell}", color=cell_numbers_color, zorder=22,
                ha='center', va="center", fontsize=fontsize, fontweight=fontweight)

    def get_cell_new_coord_in_source(self, cell, minx, miny):
        coord = self.coords[cell]
        # coords = coords - 1
        coord = coord.astype(int)
        n_coord = len(coord[0, :])
        xy = np.zeros((n_coord, 2))
        for n in np.arange(n_coord):
            # shifting the coordinates in the square size_square+1
            xy[n, 0] = coord[0, n] - minx
            xy[n, 1] = coord[1, n] - miny
        return xy

    def scale_polygon_to_source(self, poly_gon, minx, miny):
        coords = list(poly_gon.exterior.coords)
        scaled_coords = []
        for coord in coords:
            scaled_coords.append((coord[0] - minx, coord[1] - miny))
        # print(f"scaled_coords {scaled_coords}")
        return geometry.Polygon(scaled_coords)

    def get_source_profile(self, cell, tiff_movie, traces, peak_nums, spike_nums,
                           pixels_around=0, bounds=None, buffer=None, with_full_frame=False):
        """
        Return the source profile of a cell
        :param cell:
        :param pixels_around:
        :param bounds: how much padding around the cell pretty much, coordinate of the frame covering the source profile
        4 int list
        :param buffer:
        :param with_full_frame:  Average the full frame
        :return:
        """
        # print("get_source_profile")
        len_frame_x = tiff_movie[0].shape[1]
        len_frame_y = tiff_movie[0].shape[0]

        # determining the size of the square surrounding the cell
        poly_gon = self.cells_polygon[cell]
        if bounds is None:
            minx, miny, maxx, maxy = np.array(list(poly_gon.bounds)).astype(int)
        else:
            minx, miny, maxx, maxy = bounds

        if with_full_frame:
            minx = 0
            miny = 0
            maxx = len_frame_x - 1
            maxy = len_frame_y - 1
        else:
            minx = max(0, minx - pixels_around)
            miny = max(0, miny - pixels_around)
            maxx = min(len_frame_x - 1, maxx + pixels_around)
            maxy = min(len_frame_y - 1, maxy + pixels_around)

        len_x = maxx - minx + 1
        len_y = maxy - miny + 1

        # mask used in order to keep only the cells pixel
        # the mask put all pixels in the polygon, including the pixels on the exterior line to zero
        scaled_poly_gon = self.scale_polygon_to_source(poly_gon=poly_gon, minx=minx, miny=miny)
        img = PIL.Image.new('1', (len_x, len_y), 1)
        if buffer is not None:
            scaled_poly_gon = scaled_poly_gon.buffer(buffer)
        ImageDraw.Draw(img).polygon(list(scaled_poly_gon.exterior.coords), outline=0, fill=0)
        mask = np.array(img)
        # mask = np.ones((len_x, len_y))
        # cv2.fillPoly(mask, scaled_poly_gon, 0)
        # mask = mask.astype(bool)

        source_profile = np.zeros((len_y, len_x))

        # selectionning the best peak to produce the source_profile
        peaks = np.where(peak_nums[cell, :] > 0)[0]
        threshold = np.percentile(traces[cell, peaks], 95)
        selected_peaks = peaks[np.where(traces[cell, peaks] > threshold)[0]]
        # max 10 peaks, min 5 peaks
        if len(selected_peaks) > 10:
            p = 10 / len(peaks)
            threshold = np.percentile(traces[cell, peaks], (1 - p) * 100)
            selected_peaks = peaks[np.where(traces[cell, peaks] > threshold)[0]]
        elif (len(selected_peaks) < 5) and (len(peaks) > 5):
            p = 5 / len(peaks)
            threshold = np.percentile(traces[cell, peaks], (1 - p) * 100)
            selected_peaks = peaks[np.where(traces[cell, peaks] > threshold)[0]]

        # print(f"threshold {threshold}")
        # print(f"n peaks: {len(selected_peaks)}")

        onsets_frames = np.where(spike_nums[cell, :] > 0)[0]
        pos_traces = np.copy(traces)
        pos_traces += abs(np.min(traces))
        for peak in selected_peaks:
            tmp_source_profile = np.zeros((len_y, len_x))
            onsets_before_peak = np.where(onsets_frames <= peak)[0]
            if len(onsets_before_peak) == 0:
                # shouldn't arrive
                continue
            onset = onsets_frames[onsets_before_peak[-1]]
            # print(f"onset {onset}, peak {peak}")
            frames_tiff = tiff_movie[onset:peak + 1]
            for frame_index, frame_tiff in enumerate(frames_tiff):
                tmp_source_profile += (frame_tiff[miny:maxy + 1, minx:maxx + 1] * pos_traces[cell, onset + frame_index])
            # averaging
            tmp_source_profile = tmp_source_profile / (np.sum(pos_traces[cell, onset:peak + 1]))
            source_profile += tmp_source_profile
        if len(selected_peaks) > 0:
            source_profile = source_profile / len(selected_peaks)

        return source_profile, minx, miny, mask

    def get_transient_profile(self, cell, transient, tiff_movie, traces,
                              pixels_around=0, bounds=None):
        len_frame_x = tiff_movie[0].shape[1]
        len_frame_y = tiff_movie[0].shape[0]

        # determining the size of the square surrounding the cell
        if bounds is None:
            poly_gon = self.cells_polygon[cell]
            minx, miny, maxx, maxy = np.array(list(poly_gon.bounds)).astype(int)
        else:
            minx, miny, maxx, maxy = bounds

        minx = max(0, minx - pixels_around)
        miny = max(0, miny - pixels_around)
        maxx = min(len_frame_x - 1, maxx + pixels_around)
        maxy = min(len_frame_y - 1, maxy + pixels_around)

        len_x = maxx - minx + 1
        len_y = maxy - miny + 1

        transient_profile = np.zeros((len_y, len_x))
        frames_tiff = tiff_movie[transient[0]:transient[-1] + 1]
        # print(f"transient[0] {transient[0]}, transient[1] {transient[1]}")
        # now we do the weighted average
        raw_traces = np.copy(traces)
        # so the lowest value is zero
        raw_traces += abs(np.min(raw_traces))

        for frame_index, frame_tiff in enumerate(frames_tiff):
            transient_profile += (
                    frame_tiff[miny:maxy + 1, minx:maxx + 1] * raw_traces[cell, transient[0] + frame_index])
        # averaging
        transient_profile = transient_profile / (np.sum(raw_traces[cell, transient[0]:transient[-1] + 1]))

        return transient_profile, minx, miny

    def corr_between_source_and_transient(self, cell, transient, source_profile_dict, tiff_movie, traces,
                                          source_profile_corr_dict=None,
                                          pixels_around=1):
        """
        Measure the correlation (pearson) between a source and transient profile for a giveb cell
        :param cell:
        :param transient:
        :param source_profile_dict should contains cell as key, and results of get_source_profile avec values
        :param pixels_around:
        :param source_profile_corr_dict: if not None, used to save the correlation of the source profile, f
        for memory and computing proficiency
        :return:
        """
        # print('corr_between_source_and_transient')
        poly_gon = self.cells_polygon[cell]

        # Correlation test
        bounds_corr = np.array(list(poly_gon.bounds)).astype(int)
        # looking if this source has been computed before for correlation
        if (source_profile_corr_dict is not None) and (cell in source_profile_corr_dict):
            source_profile_corr, mask_source_profile = source_profile_corr_dict[cell]
        else:
            source_profile_corr, minx_corr, \
            miny_corr, mask_source_profile, xy_source = source_profile_dict[cell]
            # normalizing
            source_profile_corr = source_profile_corr - np.mean(source_profile_corr)
            # we want the mask to be at ones over the cell
            mask_source_profile = (1 - mask_source_profile).astype(bool)
            if source_profile_corr_dict is not None:
                source_profile_corr_dict[cell] = (source_profile_corr, mask_source_profile)

        transient_profile_corr, minx_corr, miny_corr = self.get_transient_profile(cell=cell,
                                                                                  transient=transient,
                                                                                  tiff_movie=tiff_movie, traces=traces,
                                                                                  pixels_around=pixels_around,
                                                                                  bounds=bounds_corr)
        transient_profile_corr = transient_profile_corr - np.mean(transient_profile_corr)

        pearson_corr, pearson_p_value = stats.pearsonr(source_profile_corr[mask_source_profile],
                                                       transient_profile_corr[mask_source_profile])

        return pearson_corr


def _angle_to_point(point, centre):
    '''calculate angle in 2-D between points and x axis'''
    delta = point - centre
    res = np.arctan(delta[1] / delta[0])
    if delta[0] < 0:
        res += np.pi
    return res


def area_of_triangle(p1, p2, p3):
    '''calculate area of any triangle given co-ordinates of the corners'''
    return np.linalg.norm(np.cross((p2 - p1), (p3 - p1))) / 2.


def convex_hull(points, smidgen=0.0075):
    '''
    from: https://stackoverflow.com/questions/17553035/draw-a-smooth-polygon-around-data-points-in-a-scatter-plot-in-matplotlib
    Calculate subset of points that make a convex hull around points
    Recursively eliminates points that lie inside two neighbouring points until only convex hull is remaining.

    :Parameters:
    points : ndarray (2 x m)
    array of points for which to find hull
    use pylab to show progress?
    smidgen : float
    offset for graphic number labels - useful values depend on your data range

    :Returns:
    hull_points : ndarray (2 x n)
    convex hull surrounding points
    '''

    n_pts = points.shape[1]
    # assert(n_pts > 5)
    centre = points.mean(1)

    angles = np.apply_along_axis(_angle_to_point, 0, points, centre)
    pts_ord = points[:, angles.argsort()]

    pts = [x[0] for x in zip(pts_ord.transpose())]
    prev_pts = len(pts) + 1
    k = 0
    while prev_pts > n_pts:
        prev_pts = n_pts
        n_pts = len(pts)
        i = -2
        while i < (n_pts - 2):
            Aij = area_of_triangle(centre, pts[i], pts[(i + 1) % n_pts])
            Ajk = area_of_triangle(centre, pts[(i + 1) % n_pts], pts[(i + 2) % n_pts])
            Aik = area_of_triangle(centre, pts[i], pts[(i + 2) % n_pts])
            if Aij + Ajk < Aik:
                del pts[i + 1]
            i += 1
            n_pts = len(pts)
        k += 1
    return np.asarray(pts)



# import tensorflow as tf

# TF_VERSION = tf.__version__
# print(f"TF_VERSION {TF_VERSION}")
#
# # depending on the TF version installed
# if TF_VERSION[0] == "2":
#     from tensorflow.keras.models import model_from_json
# else:
#     from keras.models import model_from_json

import numpy as np
from deepcinac.utils.utils import smooth_convolve, create_one_npy_file_by_frame, load_movie, \
    create_one_tiff_file_by_frame


from deepcinac.utils.cells_map_utils import CellsCoord, create_cells_coord_from_suite_2p
import PIL
import os
from abc import ABC, abstractmethod
from ScanImageTiffReader import ScanImageTiffReader
from shapely.geometry import MultiPoint, LineString


class CinacMovie(ABC):
    def __init__(self):
        self.dimensions = None
        self.n_frames = 0

    @abstractmethod
    def get_frames_section(self, frames, minx, maxx, miny, maxy):
        pass

    def get_dimensions(self):
        """
        Get x and y dimensions of the movie
        Returns: a 1d array of integers

        """
        return self.dimensions

    def get_n_frames(self):
        """
        The number of frames in the movie
        Returns: integer

        """
        return self.n_frames

    @abstractmethod
    def get_full_movie(self, normalized):
        """
        Full movie, if available a 3d array n_frames x len_y x len_x
        Args:
            normalized: bool, if True return normalized movie, False original movie if available

        Returns:

        """
        pass


class CinacDataMovie(CinacMovie):
    """
    Take the movie as a 2d array directly
    """

    def __init__(self, movie, already_normalized=False):
        super().__init__()

        self.movie = None
        if already_normalized:
            self.movie_normalized = movie
        else:
            self.movie = movie
            self.movie_normalized = movie - np.mean(movie)
            self.movie_normalized = self.movie_normalized / np.std(movie)
        self.n_frames = len(self.movie_normalized)
        self.dimensions = self.movie_normalized.shape[1:]

    def get_frames_section(self, frames, minx, maxx, miny, maxy):
        return self.movie_normalized[frames, miny:maxy + 1, minx:maxx + 1]

    def get_full_movie(self, normalized):
        """
        Full movie, if available a 3d array n_frames x len_y x len_x
        Args:
            normalized: bool, if True return normalized movie, False original movie if available

        Returns:

        """
        if normalized or (self.movie is None):
            return self.movie_normalized
        return self.movie


class CinacTiffMovie(CinacMovie):
    def __init__(self, tiff_file_name=None, tiff_movie=None):
        super().__init__()
        if tiff_movie is None:
            if tiff_file_name is None:
                raise Exception("tiff_file_name and tiff_movie can't both be set to None.")
            self.tiff_movie_normalized = load_movie(file_name=tiff_file_name,
                                                    with_normalization=True, verbose=True)
            self.tiff_movie = None
        else:
            self.tiff_movie = tiff_movie
            self.tiff_movie_normalized = tiff_movie - np.mean(tiff_movie)
            self.tiff_movie_normalized = self.tiff_movie_normalized / np.std(tiff_movie)

        self.n_frames = len(self.tiff_movie_normalized)
        self.dimensions = self.tiff_movie_normalized.shape[1:]

    def get_frames_section(self, frames, minx, maxx, miny, maxy):
        return self.tiff_movie_normalized[frames, miny:maxy + 1, minx:maxx + 1]

    def get_full_movie(self, normalized):
        """
        Full movie, if available a 3d array n_frames x len_y x len_x
        Args:
            normalized: bool, if True return normalized movie, False original movie if available

        Returns:

        """
        if normalized or (self.tiff_movie is None):
            return self.tiff_movie_normalized
        return self.tiff_movie


class CinacFileReaderMovie(CinacMovie):
    def __init__(self, cinac_file_reader, segment):
        """

        Args:
            cinac_file_reader: CinacFileReader instance
            segment: tuple of 3 int, cell, first_frame and last_frame to identify the segment
        """
        super().__init__()
        self.cinac_file_reader = cinac_file_reader
        self.segment = segment

        self.n_frames = segment[2] - segment[1] + 1

        self.dimensions = self.cinac_file_reader.get_segment_ci_movie_frames(segment=self.segment,
                                                                             frames=np.array([0])).shape[1:]

    def get_frames_section(self, frames, minx, maxx, miny, maxy):
        frames_section = self.cinac_file_reader.get_segment_ci_movie_frames(segment=self.segment, frames=frames)
        # print(f"CinacFileReaderMovie frames_section.shape {frames_section.shape}")
        coords = [minx, maxx, miny, maxy]
        # print(f"CinacFileReaderMovie coords {coords}")
        frames_section = frames_section[:, miny:maxy + 1, minx:maxx + 1]
        # print(f"CinacFileReaderMovie frames_section.shape bis {frames_section.shape}")
        return frames_section

    def get_full_movie(self, normalized):
        """
        Full movie, if available a 3d array n_frames x len_y x len_x
        Args:
            normalized: bool, if True return normalized movie, False original movie if available

        Returns:

        """
        print("### CinacFileReaderMovie get_full_movie()")
        # not keeping the full movie in memory so that the garbage collector could free the memory
        # if the movie is not used anymore elsewhere, but means the file will be read everytime the function
        # get_full_movie() will be called
        ci_movie = self.cinac_file_reader.get_segment_ci_movie(segment=self.segment)
        if normalized:
            return (ci_movie - np.mean(ci_movie)) / np.std(ci_movie)
        return ci_movie


class CinacSplitedTiffMovie(CinacMovie):
    """
    Used if tiff movie has been splitted as as many tiff files as frames in the movie
    """

    def __init__(self, identifier, tiffs_dirname, already_normalized=False, tiff_file_name=None, tiff_movie=None):
        """

        Args:
            identifier: string
            tiffs_dirname: dirname of where to save the tiffs created
            tiff_file_name: string, if not None, will create a tiff by frame from this tiff movie if not already done.
            Will be saved in tiffs_dirname. If tiff_movie is not None, the tiff movie will be chosen.
            tiff_movie: numpy 3d array f not None, will create a tiff by frame from this tiff movie if not already done.
            Will be saved in tiffs_dirname.
            already_normalized: indicate if the movie if already normalized (valid if tiff_file_name or tiff_movie
            is not None. Otherwise, normalization will be done if a mean and std files are found in the tiffs directory
        """
        super().__init__()
        self.tiffs_dirname = tiffs_dirname
        self.identifier = identifier

        self.tiff_movie = tiff_movie
        self.movie_normalized = None

        # first if tiff_file_name or tiff_movie are not None, it means we should try to create the SplittedTiffVersion
        # if they doen't exist yet
        if tiff_file_name is not None or tiff_movie is not None:
            movies_to_split = dict()
            if tiff_movie is not None:
                movies_to_split[identifier] = tiff_movie
            else:
                movies_to_split[identifier] = tiff_file_name
            # won't do anything if the tiffs already have been created
            create_one_tiff_file_by_frame(movies_to_split=movies_to_split,
                                              results_path=tiffs_dirname,
                                              without_mean_std_files=already_normalized,
                                              verbose=2)

        self.tiff_movie_mean = None
        self.tiff_movie_std = None

        file_names = []

        self.dir_to_explore = None
        dir_names = []
        # first we identify which directory contains the tiffs
        # look for filenames in the fisrst directory, if we don't break, it will go through all directories
        for (dirpath, dirnames, local_filenames) in os.walk(os.path.join(self.tiffs_dirname)):
            dir_names.extend(dirnames)
            break
        for dir_name in dir_names:
            if dir_name in self.identifier.lower():
                self.dir_to_explore = dir_name
                break
        if self.dir_to_explore is None:
            raise Exception(f"No directory with tiffs for {self.identifier} has been found in {self.tiffs_dirname}")

        # look for filenames in the fisrst directory, if we don't break, it will go through all directories
        for (dirpath, dirnames, local_filenames) in os.walk(os.path.join(self.tiffs_dirname, self.dir_to_explore)):
            file_names.extend(local_filenames)
            break

        # we loop in the directory filenames in order to set the mean and std of the movie and count how many frames
        # are in the movie (according they have all be saved correctly)
        # if mean and std are not present, it means the movie is already normalized
        self.n_frames = 0
        for file_name in file_names:
            if file_name.endswith("mean.npy"):
                mean_value = np.load(os.path.join(self.tiffs_dirname, self.dir_to_explore, file_name))
                self.tiff_movie_mean = mean_value
            elif file_name.endswith("std.npy"):
                std_value = np.load(os.path.join(self.tiffs_dirname, self.dir_to_explore, "std.npy"))
                self.tiff_movie_std = std_value
            # elif file_name.endswith("tiff") or file_name.endswith("tif"):
            elif file_name.endswith("tiff"):
                self.n_frames += 1
                if self.dimensions is None:
                    # im = np.load(os.path.join(self.tiffs_dirname, self.identifier, file_name))
                    im = PIL.Image.open(os.path.join(self.tiffs_dirname, self.dir_to_explore, file_name))
                    im = np.array(im)
                    self.dimensions = im.shape

    def get_frames_section(self, frames_indices, minx, maxx, miny, maxy):
        """
        get section of given frames from the calcium imaging movie
        Args:
            frames_indices: numpy array of integers, representing the frame's indices to select
            minx: integer, min x coordinate
            maxx: integer, max x coordinate
            miny: integer, min y coordinate
            maxy: integer, max y coordinate

        Returns:

        """

        frames_section = np.zeros((len(frames_indices), maxy - miny + 1, maxx - minx + 1))
        for frame_index, frame in enumerate(frames_indices):
            # im = np.load(os.path.join(self.tiffs_dirname, self.identifier, f"{frame}.npy"))
            try:
                im = ScanImageTiffReader(os.path.join(self.tiffs_dirname,
                                                      self.dir_to_explore, f"{frame}.tiff")).data()
            except Exception as e:
                im = PIL.Image.open(os.path.join(self.tiffs_dirname,
                                                 self.dir_to_explore, f"{frame}.tiff"))
                im = np.array(im)
            frames_section[frame_index] = im[miny:maxy + 1, minx:maxx + 1]
        # normalizing using the mean and std from the whole movie
        if self.tiff_movie_mean is not None:
            frames_section = (frames_section - self.tiff_movie_mean) / self.tiff_movie_std
        # if None, then we suppose the movie is already normalized
        return frames_section

    def get_full_movie(self, normalized):
        """
        Full movie, if available a 3d array n_frames x len_y x len_x
        Args:
            normalized: bool, if True return normalized movie, False original movie if available

        Returns:

        """
        print("#### get_full_movie() in CinacSplitedTiffMovie")
        if self.tiff_movie is None:
            tiff_movie = np.zeros((self.n_frames, self.dimensions[0], self.dimensions[1]), dtype="uint16")
            for frame_index in np.arange(self.n_frames):
                try:
                    im = ScanImageTiffReader(os.path.join(self.tiffs_dirname,
                                                          self.dir_to_explore, f"{frame_index}.tiff")).data()
                except Exception as e:
                    im = PIL.Image.open(os.path.join(self.tiffs_dirname,
                                                     self.dir_to_explore, f"{frame_index}.tiff"))
                    im = np.array(im)
                tiff_movie[frame_index] = im

            self.tiff_movie = tiff_movie

        if normalized:
            if self.movie_normalized is None:
                if self.tiff_movie_mean is None:
                    self.movie_normalized = self.tiff_movie
                else:
                    self.movie_normalized = self.tiff_movie - self.tiff_movie_mean
                    self.movie_normalized = self.movie_normalized / self.tiff_movie_std
            return self.tiff_movie_normalized
        return self.tiff_movie


class CinacSplitedNpyMovie(CinacMovie):
    """
    Used if tiff movie has been splitted as as many tiff files as frames in the movie
    """

    def __init__(self, identifier, tiffs_dirname, already_normalized=False, tiff_file_name=None, tiff_movie=None):
        """

        Args:
            identifier: string
            tiffs_dirname: dirname of where to save the tiffs created
            tiff_file_name: string, if not None, will create a tiff by frame from this tiff movie if not already done.
            Will be saved in tiffs_dirname. If tiff_movie is not None, the tiff movie will be chosen.
            tiff_movie: numpy 3d array f not None, will create a tiff by frame from this tiff movie if not already done.
            Will be saved in tiffs_dirname.
            already_normalized: indicate if the movie if already normalized (valid if tiff_file_name or tiff_movie
            is not None. Otherwise, normalization will be done if a mean and std files are found in the tiffs directory
        """
        super().__init__()
        self.tiffs_dirname = tiffs_dirname
        self.identifier = identifier

        self.tiff_movie = tiff_movie
        self.movie_normalized = None

        # first if tiff_file_name or tiff_movie are not None, it means we should try to create the SplittedTiffVersion
        # if they doen't exist yet
        if tiff_file_name is not None or tiff_movie is not None:
            movies_to_split = dict()
            if tiff_movie is not None:
                movies_to_split[identifier] = tiff_movie
            else:
                movies_to_split[identifier] = tiff_file_name
            # won't do anything if the tiffs already have been created
            # used to be create_one_tiff_file_by_frame
            create_one_npy_file_by_frame(movies_to_split=movies_to_split,
                                              results_path=tiffs_dirname,
                                              without_mean_std_files=already_normalized,
                                              verbose=2)

        self.tiff_movie_mean = None
        self.tiff_movie_std = None

        file_names = []

        # look for filenames in the fisrst directory, if we don't break, it will go through all directories
        for (dirpath, dirnames, local_filenames) in os.walk(os.path.join(self.tiffs_dirname, self.identifier)):
            file_names.extend(local_filenames)
            break

        # we loop in the directory filenames in order to set the mean and std of the movie and count how many frames
        # are in the movie (according they have all be saved correctly)
        # if mean and std are not present, it means the movie is already normalized
        self.n_frames = 0
        for file_name in file_names:
            if file_name.endswith("mean.npy"):
                mean_value = np.load(os.path.join(self.tiffs_dirname, self.identifier, file_name))
                self.tiff_movie_mean = mean_value
            elif file_name.endswith("std.npy"):
                std_value = np.load(os.path.join(self.tiffs_dirname, self.identifier, "std.npy"))
                self.tiff_movie_std = std_value
            # elif file_name.endswith("tiff") or file_name.endswith("tif"):
            elif file_name.endswith("npy"):
                self.n_frames += 1
                if self.dimensions is None:
                    im = np.load(os.path.join(self.tiffs_dirname, self.identifier, file_name))
                    # im = PIL.Image.open(os.path.join(self.tiffs_dirname, self.identifier, file_name))
                    # im = np.array(im)
                    self.dimensions = im.shape

    def get_frames_section(self, frames_indices, minx, maxx, miny, maxy):
        """
        get section of given frames from the calcium imaging movie
        Args:
            frames_indices: numpy array of integers, representing the frame's indices to select
            minx: integer, min x coordinate
            maxx: integer, max x coordinate
            miny: integer, min y coordinate
            maxy: integer, max y coordinate

        Returns:

        """

        frames_section = np.zeros((len(frames_indices), maxy - miny + 1, maxx - minx + 1))
        for frame_index, frame in enumerate(frames_indices):
            im = np.load(os.path.join(self.tiffs_dirname, self.identifier, f"{frame}.npy"))
            # try:
            #     im = ScanImageTiffReader(os.path.join(self.tiffs_dirname,
            #                                           self.identifier, f"{frame}.tiff")).data()
            # except Exception as e:
            #     im = PIL.Image.open(os.path.join(self.tiffs_dirname,
            #                                      self.identifier, f"{frame}.tiff"))
            #     im = np.array(im)
            frames_section[frame_index] = im[miny:maxy + 1, minx:maxx + 1]
        # normalizing using the mean and std from the whole movie
        if self.tiff_movie_mean is not None:
            frames_section = (frames_section - self.tiff_movie_mean) / self.tiff_movie_std
        # if None, then we suppose the movie is already normalized
        return frames_section

    def get_full_movie(self, normalized):
        """
        Full movie, if available a 3d array n_frames x len_y x len_x
        Args:
            normalized: bool, if True return normalized movie, False original movie if available

        Returns:

        """
        print("#### get_full_movie() in CinacSplitedNpyMovie")
        if self.tiff_movie is None:
            tiff_movie = np.zeros((self.n_frames, self.dimensions[0], self.dimensions[1]), dtype="uint16")
            for frame_index in np.arange(self.n_frames):
                try:
                    im = ScanImageTiffReader(os.path.join(self.tiffs_dirname,
                                                          self.identifier, f"{frame_index}.tiff")).data()
                except Exception as e:
                    im = PIL.Image.open(os.path.join(self.tiffs_dirname,
                                                     self.identifier, f"{frame_index}.tiff"))
                    im = np.array(im)
                tiff_movie[frame_index] = im

            self.tiff_movie = tiff_movie

        if normalized:
            if self.movie_normalized is None:
                if self.tiff_movie_mean is None:
                    self.movie_normalized = self.tiff_movie
                else:
                    self.movie_normalized = self.tiff_movie - self.tiff_movie_mean
                    self.movie_normalized = self.movie_normalized / self.tiff_movie_std
            return self.tiff_movie_normalized
        return self.tiff_movie




def get_cinac_movie_from_cinac_file_reader(cinac_file_reader):
    """

    Args:
        cinac_file_reader: CinacFileReader instance

    Returns:

    """
    ci_movie_file_name = cinac_file_reader.get_ci_movie_file_name()
    if ci_movie_file_name is None or (not os.path.exists(ci_movie_file_name)):
        raise Exception(f"ci_movie_file_name in cinac_file should be valid: {ci_movie_file_name}")

    ci_movie = load_movie(file_name=ci_movie_file_name, both_instances=False,
                          with_normalization=False, verbose=True)
    cinac_movie = CinacDataMovie(movie=ci_movie, already_normalized=False)
    return cinac_movie


def create_cinac_recording_from_cinac_file_segment(identifier, cinac_file_reader, segment):
    """

    Args:
        identifier: cinac_recording identifier
        cinac_file_reader: CinacFileReaderInstance
        cinac_movie: CinacMovie instance
        segment: tuple of 3 int (cell, first_frame, last_frame

    Returns:

    """
    # cinac_file_reader = CinacFileReader(file_name=cinac_file_name)
    # segments_list = cinac_file_reader.get_all_segments()
    cinac_movie = CinacFileReaderMovie(cinac_file_reader=cinac_file_reader,
                                       segment=segment)
    cinac_recording = CinacRecording(identifier)
    cinac_recording.set_movie(cinac_movie)
    # smooth_traces is already saved in z_score normalized format
    cinac_recording.smooth_traces = cinac_file_reader.get_segment_smooth_traces(segment=segment)
    cinac_recording.smooth_traces_z_score = cinac_recording.smooth_traces
    # ci_movie = cinac_movie.get_full_movie(normalized=True)
    coords_data = cinac_file_reader.get_segment_cells_contour(segment)
    coords_data = [np.vstack((coord_data[0], coord_data[1])) for coord_data in coords_data]
    # coord_obj = CellsCoord(coords=coords_data, pixel_masks=None, nb_lines=ci_movie.shape[1],
    #                        nb_col=ci_movie.shape[2],
    #                        from_matlab=False, invert_xy_coord=False)
    # cinac_recording.coord_obj = coord_obj
    invalid_cells = cinac_file_reader.get_segment_invalid_cells(segment=segment)
    # invalid_cells binary array same length as the number of cell, 1 if the cell is invalid
    # invalid cells allows to remove contours, so the classifier don't take it in consideration
    # this work because the cell we are interesting with is the cell 0 and should never be invalid
    if np.sum(invalid_cells) > 0:
        new_coords_data = []
        for cell_index, cell_coord in enumerate(coords_data):
            if invalid_cells[cell_index] > 0:
                continue
            new_coords_data.append(cell_coord)
        coords_data = new_coords_data

    cinac_recording.set_rois_2d_array(coord=coords_data, from_matlab=False)

    cinac_recording.cell_type = cinac_file_reader.get_segment_cell_type(segment=segment)
    return cinac_recording

class CinacRecording:
    # TODO: add fields that will contains the ground truth

    def __init__(self, identifier):
        self.cinac_movie = None
        self.identifier = identifier
        self.coord_obj = None
        # string defining the cell type of the recording, if known means it is unknown
        self.cell_type = None
        self.raw_traces = None
        self.raw_traces_z_score = None
        self.smooth_traces = None
        self.smooth_traces_z_score = None

    def _build_traces(self):
        self.raw_traces = self.coord_obj.build_raw_traces_from_movie(movie=
                                                                     self.cinac_movie.get_full_movie(normalized=False))
        self.smooth_traces = np.copy(self.raw_traces)
        windows = ['hanning', 'hamming', 'bartlett', 'blackman']
        i_w = 1
        window_length = 7  # 11
        for i in np.arange(self.smooth_traces.shape[0]):
            smooth_signal = smooth_convolve(x=self.smooth_traces[i], window_len=window_length,
                                            window=windows[i_w])
            beg = (window_length - 1) // 2
            self.smooth_traces[i] = smooth_signal[beg:-beg]
        self.smooth_traces_z_score = np.zeros(self.smooth_traces.shape)
        self.raw_traces_z_score = np.zeros(self.raw_traces.shape)
        for i in np.arange(len(self.raw_traces)):
            # -2 so smooth smooth_traces will be displayed under the raw smooth_traces
            self.smooth_traces_z_score[i, :] = ((self.smooth_traces[i, :] - np.mean(self.smooth_traces[i, :])) / np.std(
                self.smooth_traces[i, :])) - 2
            self.raw_traces_z_score[i, :] = (self.raw_traces[i, :] - np.mean(self.raw_traces[i, :])) \
                                            / np.std(self.raw_traces[i, :])

    def get_raw_traces(self, normalized):
        """

        Args:
            normalized: if True, get the raw_traces normalized

        Returns:

        """
        if self.raw_traces is None:
            self._build_traces()

        if normalized:
            return self.raw_traces_z_score
        else:
            return self.raw_traces

    def get_smooth_traces(self, normalized):
        """

        Args:
            normalized: if True, get the raw_traces normalized

        Returns:

        """
        if self.smooth_traces is None:
            self._build_traces()

        if normalized:
            return self.smooth_traces_z_score
        else:
            return self.smooth_traces

    def set_movie(self, cinac_movie):
        """
        Set the instance of CinacMovie, that will be used to get the frames given to the network
        Args:
            cinac_movie:

        Returns:

        """
        self.cinac_movie = cinac_movie

    def set_rois_from_suite_2p(self, is_cell_file_name, stat_file_name):
        """

        Args:
            is_cell_file_name: path and file_name of the file iscell.npy produce by suite2p segmentation process
            stat_file_name: path and file_name of the file stat.npy produce by suite2p segmentation process

        Returns:

        """
        if self.cinac_movie is None:
            raise Exception(f"cinac_movie should be set using the method set_movie() before setting the Rois")
        pass

        self.coord_obj = create_cells_coord_from_suite_2p(is_cell_file_name=is_cell_file_name,
                                                          stat_file_name=stat_file_name,
                                                          movie_dimensions=self.cinac_movie.get_dimensions())

    def set_rois_2d_array(self, coord, from_matlab):
        """

        Args:
            coord: numpy array of 2d, first dimension of length 2 (x and y) and 2nd dimension of length the number of
            cells. Could also be a list of lists or tuples of 2 integers
            from_matlab: Indicate if the data has been computed by matlab, then 1 will be removed to the coordinates
            so it starts at zero.

        Returns:

        """
        if self.cinac_movie is None:
            raise Exception(f"cinac_movie should be set using the method set_movie() before setting the Rois")
        dimensions = self.cinac_movie.get_dimensions()
        self.coord_obj = CellsCoord(coords=coord, nb_lines=dimensions[0], nb_col=dimensions[1], from_matlab=from_matlab)

    def set_rois_from_nwb(self, nwb_data, name_module, name_segmentation, name_seg_plane):
        """

        Args:
            nwb_data: nwb object instance
            name_module: Name of the module to find segmentation. Will be used this way: nwb_data.modules[name_module]
                Ex: name_module = 'ophys'
            name_segmentation: Name of the segmentation in which find the plane segmentation.
                Used this way:get_plane_segmentation(name_segmentation)
                Ex: name_segmentation = 'segmentation_suite2p'
            name_seg_plane: Name of the segmentation plane in which to find the ROIs data
            Used this way: mod[name_segmentation]get_plane_segmentation(name_seq_plane)
                Ex: name_segmentation = 'my_plane_seg'

        Returns:

        """
        if self.cinac_movie is None:
            raise Exception(f"cinac_movie should be set using the method set_movie() before setting the Rois")
        mod = nwb_data.modules[name_module]
        plane_seg = mod[name_segmentation].get_plane_segmentation(name_seg_plane)

        if 'pixel_mask' not in plane_seg:
            raise Exception("'pixel_mask' has to exist in plane_segmentation in order to create ROIs")

        self.set_rois_using_pixel_mask(pixel_masks=plane_seg['pixel_mask'])

    def set_rois_using_pixel_mask(self, pixel_masks):
        """

        Args:
            pixel_masks: list of list of 2 integers representing for each cell all the pixels that belongs to the cell

        Returns:

        """
        if self.cinac_movie is None:
            raise Exception(f"cinac_movie should be set using the method set_movie() before setting the Rois")

        # TODO: use pixel_mask instead of using the coord of the contour of the cell
        #  means changing the way coord_cell works
        coord_list = []
        for cell in np.arange(len(pixel_masks)):
            pixels_coord = pixel_masks[cell]
            list_points_coord = [(pix[0], pix[1]) for pix in pixels_coord]
            convex_hull = MultiPoint(list_points_coord).convex_hull
            if isinstance(convex_hull, LineString):
                coord_shapely = MultiPoint(list_points_coord).convex_hull.coords
            else:
                coord_shapely = MultiPoint(list_points_coord).convex_hull.exterior.coords
            coord_list.append(np.array(coord_shapely).transpose())

        dimensions = self.cinac_movie.get_dimensions()
        self.coord_obj = CellsCoord(coords=coord_list, nb_lines=dimensions[0], nb_col=dimensions[1],
                                    from_matlab=False)

    def get_n_frames(self):
        """
        Return the number of frames in the movie
        Returns:

        """
        return self.cinac_movie.get_n_frames()

    def get_n_cells(self):
        return self.coord_obj.n_cells

    def get_source_profile_frames(self, frames_indices, coords):
        """
        Return frames section based on the indices of the frames and the coordinates of the corners of the section
        Args:
            frames_indices: array of integers
            coords: tuple of 4 integers: (minx, maxx, miny, maxy)

        Returns: A numpy array of dimensions len(frames_indices) * (maxy - miny + 1) * (maxx - minx + 1)

        """

        frames_section = self.cinac_movie.get_frames_section(frames_indices, *coords)

        return frames_section













# -*- coding: utf-8 -*-
"""
###DEMO of deepCINAC###

We're going to guide you on how using deepCINAC to train either a cell type or an activity classifier from your
calcium imaging data. Most of the code is similar and in both case you will need to produce labeled data using our GUI.

This notebook has been conceived in order to be run on google colab.
A [python](https://gitlab.com/cossartlab/deepcinac/tree/master/demos/general/demo_deep_cinac_cell_type_training.py) file
is available to be run localy.

Here is a link to our [gitlab page](https://gitlab.com/cossartlab/deepcinac) for more information about our package.

So far, to run this code, you will need some calcium imaging data to work on (in tiff format) and some segmentation data
(ROIs) indicating the contours or pixels that compose your cells
(compatible format are Caiman, suite2p, Fiji or NWB outputs). You will need to open those data using our GUI
(see the [tutorial](https://deepcinac.readthedocs.io/en/latest/tutorial_gui.html))
in order to annotate it (either with the cell type and/or activity) and to produce .cinac files
that will be used to train the classifier.

"""

"""A few packages need to be installed"""

from deepcinac.cinac_model import *
from deepcinac.cinac_predictor import *

if __name__ == '__main__':
    """Change the paths according to your environment"""
    root_path = "/content/deepcinac/"

    data_path = os.path.join(root_path, "data/")
    results_path = os.path.join(root_path, "results")
    time_str = datetime.now().strftime("%Y_%m_%d.%H-%M-%S")
    results_path = os.path.join(results_path, time_str)
    os.mkdir(results_path)

    # indicate if you want to train a classifier to infer neuronal activity or cell type.
    train_activity_classifier = True

    if train_activity_classifier:
        cinac_dir_name = os.path.join(data_path, "cinac_ground_truth/for_training")
    else:
        cinac_dir_name = os.path.join(data_path,
                                      "cinac_cell_type_ground_truth/for_training")

    if not train_activity_classifier:
        """TO PREPARE CELL TYPE CLASSIFIER"""

        """
          To start training from a full saved model from another training, it is necessary to:
          - During the previous training, to put as argument: save_only_the_weitghs=False
          - then specify the .h5 containing the model using partly_trained_model
          - finally setting the learning rate so it is the same as the last epoch trained, using learning_rate_start
        """

        partly_trained_model = os.path.join(root_path, "")
        # yaml_file = os.path.join(data_path, "cell_type_categories_default.yaml")
        cinac_model = CinacModel(results_path=results_path, n_epochs=10, verbose=1,
                                 batch_size=4,
                                 cell_type_classifier_mode=True,
                                 max_width=10, max_height=10,
                                 window_len=1000, max_n_transformations=1,
                                 n_windows_len_to_keep_by_cell=2,
                                 conv_filters=(32, 32, 64, 64),
                                 lstm_layers_size=[32], bin_lstm_size=32,
                                 overlap_value=0.5,
                                 # frames_to_avoid_for_cell_type=[2500, 5000, 7500, 10000],
                                 # partly_trained_model=partly_trained_model,
                                 #  learning_rate_start = 0.001,
                                 save_only_the_weitghs=False
                                 # cell_type_categories_yaml_file=yaml_file
                                 )
    else:
        """TO PREPARE CELL ACTIVITY CLASSIFIER"""

        """
            To start training from a full saved model from another training, it is necessary to:
            - During the previous training, to put as argument: save_only_the_weitghs=False
            - then specify the .h5 containing the model using partly_trained_model
            - finally setting the learning rate so it is the same as the last epoch trained, using learning_rate_start
        """

        # partly_trained_model = os.path.join(root_path, "")
        cinac_model = CinacModel(results_path=results_path, n_epochs=20, verbose=1,
                                 batch_size=8,
                                 cell_type_classifier_mode=False,
                                 window_len=100,
                                 lstm_layers_size=[128, 256], bin_lstm_size=256,
                                 conv_filters=(64, 64, 128, 128),
                                 save_only_the_weitghs=False
                                 # partly_trained_model=partly_trained_model,
                                 #  learning_rate_start = 0.001,
                                 )

    # adding all cinac file in a directory
    cinac_model.add_input_data_from_dir(dir_name=cinac_dir_name, verbose=1)

    # cinac files can also be added individually or as a list
    # using the method add_input_data()
    # cinac_model.add_input_data(cinac_file_names=path_to_file)

    cinac_model.prepare_model(verbose=1)

    """TO START THE TRAINING"""

    cinac_model.fit()



import tensorflow as tf

TF_VERSION = tf.__version__
# print(f"TF_VERSION {TF_VERSION}")

# depending on the TF version installed
if TF_VERSION[0] == "2":
    import tensorflow.keras as keras
else:
    import keras

import numpy as np
from deepcinac.utils.utils import horizontal_flip, vertical_flip, v_h_flip, shift_movie, rotate_movie
from deepcinac.utils.utils import get_source_profile_param


class MoviePatchGenerator:
    """
    Used to generate movie patches, that will be produce for training data during each mini-batch.
    This is an abstract classes that need to have heritage.
    The function generate_movies_from_metadata will be used to produced those movie patches, the number
    vary depending on the class instantiated
    """

    def __init__(self, window_len, max_width, max_height, using_multi_class, cell_type_classifier_mode):
        self.window_len = window_len
        self.max_width = max_width
        self.max_height = max_height
        self.using_multi_class = using_multi_class
        self.cell_type_classifier_mode = cell_type_classifier_mode

    # self.n_inputs shouldn't be changed
    def get_nb_inputs(self):
        return self.n_inputs

    def generate_movies_from_metadata(self, movie_data_list, memory_dict, with_labels=True):
        pass


# TODO: See to add more MoviePatchGenerator versions

class MoviePatchGeneratorForCellType(MoviePatchGenerator):
    def __init__(self, window_len, max_width, max_height, pixels_around,
                 buffer, using_multi_class, cell_type_classifier_mode, with_all_pixels=False):
        super().__init__(window_len=window_len, max_width=max_width, max_height=max_height,
                         using_multi_class=using_multi_class, cell_type_classifier_mode=cell_type_classifier_mode)
        self.pixels_around = pixels_around
        self.buffer = buffer
        self.n_inputs = 1
        self.with_all_pixels = with_all_pixels
        if with_all_pixels:
            self.n_inputs += 1

    def generate_movies_from_metadata(self, movie_data_list, memory_dict=None, with_labels=True):
        """

        Args:
            movie_data_list: list of MoviePatchData instances
            memory_dict:
            with_labels:

        Returns:

        """
        # print(f"Start generate_movies_from_metadata() in cinac_movie_patch.py")
        source_profiles_dict = memory_dict
        if source_profiles_dict is None:
            source_profiles_dict = dict()
        batch_size = len(movie_data_list)
        if with_labels:
            if self.using_multi_class <= 1:
                if self.cell_type_classifier_mode:
                    labels = np.zeros((batch_size, 1), dtype="uint8")
                else:
                    labels = np.zeros((batch_size, self.window_len), dtype="uint8")
            else:
                if self.cell_type_classifier_mode:
                    labels = np.zeros((batch_size, self.using_multi_class), dtype="uint8")
                else:
                    labels = np.zeros((batch_size, self.window_len, self.using_multi_class), dtype="uint8")

        inputs_dict = dict()
        for input_index in np.arange(self.n_inputs):
            inputs_dict[f"input_{input_index}"] = np.zeros((batch_size, self.window_len, self.max_height,
                                                            self.max_width, 1))

        # Generate data
        for index_batch, movie_data in enumerate(movie_data_list):
            cinac_recording = movie_data.cinac_recording
            cell = movie_data.cell
            frame_index = movie_data.index_movie
            augmentation_fct = movie_data.data_augmentation_fct

            # now we generate the source profile of the cells for those frames and retrieve it if it has
            # already been generated
            # src_profile_key = cinac_recording.identifier + str(cell)
            # if src_profile_key in source_profiles_dict:
            #     mask_source_profiles, coords = source_profiles_dict[src_profile_key]
            # else:
            mask_source_profile, coords = \
                get_source_profile_param(cell=cell, movie_dimensions=cinac_recording.cinac_movie.get_dimensions(),
                                         coord_obj=cinac_recording.coord_obj, pixels_around=self.pixels_around,
                                         buffer=self.buffer,
                                         max_width=self.max_width, max_height=self.max_height,
                                         with_all_masks=False)
            # source_profiles_dict[src_profile_key] = [mask_source_profiles, coords]

            frames = np.arange(frame_index, frame_index + self.window_len)
            if with_labels:
                labels[index_batch] = movie_data.get_labels(using_multi_class=self.using_multi_class)
            # now adding the movie of those frames in this sliding_window
            # Takes around 0.65 sec on a macbook pro
            source_profile_frames = cinac_recording.get_source_profile_frames(frames_indices=frames, coords=coords)

            input_index = 0

            # then we compute the frame with just the mask of each cell (the main one (with input_0 index) and the ones
            # that overlaps)
            # mask_source_profiles_keys = np.array(list(mask_source_profiles.keys()))

            # mask_for_all_cells = np.zeros((source_profile_frames.shape[1], source_profile_frames.shape[2]),
            #                               dtype="int8")

            source_profile_frames_masked = np.copy(source_profile_frames)
            source_profile_frames_masked[:, mask_source_profile] = 0

            # doing augmentation if the function exists
            if augmentation_fct is not None:
                source_profile_frames_masked = augmentation_fct(source_profile_frames_masked)

            # then we fit it the frame use by the network, padding the surrounding by zero if necessary
            profile_fit_masked = np.zeros((len(frames), self.max_height, self.max_width))
            # we center the source profile
            y_coord = (profile_fit_masked.shape[1] - source_profile_frames.shape[1]) // 2
            x_coord = (profile_fit_masked.shape[2] - source_profile_frames.shape[2]) // 2
            profile_fit_masked[:, y_coord:source_profile_frames.shape[1] + y_coord,
            x_coord:source_profile_frames.shape[2] + x_coord] = \
                source_profile_frames_masked

            profile_fit_masked = profile_fit_masked.reshape((profile_fit_masked.shape[0],
                                                             profile_fit_masked.shape[1],
                                                             profile_fit_masked.shape[2], 1))

            inputs_dict[f"input_{input_index}"][index_batch] = profile_fit_masked


            input_index += 1

            if self.with_all_pixels:
                # doing augmentation if the function exists
                if augmentation_fct is not None:
                    # return new frames, doesn't change the original
                    source_profile_frames = augmentation_fct(source_profile_frames)
                # then we fit it the frame use by the network, padding the surrounding by zero if necessary
                profile_fit = np.zeros((len(frames), self.max_height, self.max_width))
                # we center the source profile
                y_coord = (profile_fit.shape[1] - source_profile_frames.shape[1]) // 2
                x_coord = (profile_fit.shape[2] - source_profile_frames.shape[2]) // 2
                profile_fit[:, y_coord:source_profile_frames.shape[1] + y_coord,
                x_coord:source_profile_frames.shape[2] + x_coord] = \
                    source_profile_frames

                profile_fit = profile_fit.reshape((profile_fit.shape[0], profile_fit.shape[1], profile_fit.shape[2], 1))
                data = inputs_dict[f"input_{input_index}"]
                data[index_batch] = profile_fit
                input_index += 1
        # print(f"End generate_movies_from_metadata() in cinac_movie_patch.py")

        if with_labels:
            return inputs_dict, labels
        else:
            return inputs_dict

    def __str__(self):
        bonus_str = ""
        if self.with_all_pixels:
            bonus_str = "+ one with all pixels"
        return f"{self.n_inputs} inputs. Main cell mask {bonus_str}"

class MoviePatchGeneratorMaskedVersions(MoviePatchGenerator):
    """
    Will generate one input being the masked cell (the one we focus on), the second input
    would be the whole patch without neuorpil and the main cell, the last inpu if with_neuropil_mask is True
    would be just the neuropil without the pixels in the cells
    """

    def __init__(self, window_len, max_width, max_height, pixels_around,
                 buffer, with_neuropil_mask, using_multi_class, cell_type_classifier_mode):
        super().__init__(window_len=window_len, max_width=max_width, max_height=max_height,
                         using_multi_class=using_multi_class, cell_type_classifier_mode=cell_type_classifier_mode)
        self.pixels_around = pixels_around
        self.buffer = buffer
        self.with_neuropil_mask = with_neuropil_mask
        self.n_inputs = 2
        if with_neuropil_mask:
            self.n_inputs += 1

    def generate_movies_from_metadata(self, movie_data_list, memory_dict=None, with_labels=True):
        """

        Args:
            movie_data_list: list of MoviePatchData instances
            memory_dict:
            with_labels:

        Returns:

        """
        # print(f"Start generate_movies_from_metadata() in cinac_movie_patch.py")
        source_profiles_dict = memory_dict
        if source_profiles_dict is None:
            source_profiles_dict = dict()
        batch_size = len(movie_data_list)
        if with_labels:
            if self.using_multi_class <= 1:
                if self.cell_type_classifier_mode:
                    labels = np.zeros((batch_size, 1), dtype="uint8")
                else:
                    labels = np.zeros((batch_size, self.window_len), dtype="uint8")
            else:
                if self.cell_type_classifier_mode:
                    labels = np.zeros((batch_size, 1, self.using_multi_class), dtype="uint8")
                else:
                    labels = np.zeros((batch_size, self.window_len, self.using_multi_class), dtype="uint8")

        # if there are no overlaping cells, we'll give empty frames as inputs (with pixels to zero)
        inputs_dict = dict()
        for input_index in np.arange(self.n_inputs):
            inputs_dict[f"input_{input_index}"] = np.zeros((batch_size, self.window_len, self.max_height,
                                                            self.max_width, 1))

        # Generate data
        for index_batch, movie_data in enumerate(movie_data_list):
            cinac_recording = movie_data.cinac_recording
            cell = movie_data.cell
            frame_index = movie_data.index_movie
            augmentation_fct = movie_data.data_augmentation_fct

            # now we generate the source profile of the cells for those frames and retrieve it if it has
            # already been generated
            # src_profile_key = cinac_recording.identifier + str(cell)
            # if src_profile_key in source_profiles_dict:
            #     mask_source_profiles, coords = source_profiles_dict[src_profile_key]
            # else:
            # print(f"In cinac_movie_patch.py {cinac_recording.identifier} movie_dimensions "
            #       f"{cinac_recording.cinac_movie.get_dimensions()}, max_width {self.max_width}, max_height {self.max_height}")
            mask_source_profiles, coords = \
                get_source_profile_param(cell=cell, movie_dimensions=cinac_recording.cinac_movie.get_dimensions(),
                                         coord_obj=cinac_recording.coord_obj, pixels_around=self.pixels_around,
                                         buffer=self.buffer,
                                         max_width=self.max_width, max_height=self.max_height,
                                         with_all_masks=True)
            # source_profiles_dict[src_profile_key] = [mask_source_profiles, coords]

            frames = np.arange(frame_index, frame_index + self.window_len)
            if with_labels:
                labels[index_batch] = movie_data.get_labels(using_multi_class=self.using_multi_class)
            # now adding the movie of those frames in this sliding_window
            # Takes around 0.65 sec on a macbook pro
            source_profile_frames = cinac_recording.get_source_profile_frames(frames_indices=frames, coords=coords)

            # TMP plotting
            # for index_source_prof in np.arange(0, source_profile_frames.shape[0], 10):
            #     # labels = movie_data.get_labels(using_multi_class=self.using_multi_class)
            #     # print(f"labels.shape {labels.shape}")
            #     print(f"Cell {cell} {index_source_prof} / {source_profile_frames.shape[0] - 1} from {cinac_recording.identifier}: "
            #           f"min {np.min(source_profile_frames[index_source_prof])}, "
            #           f"max {np.max(source_profile_frames[index_source_prof])}, "
            #           f"mean {np.mean(source_profile_frames[index_source_prof])}, "
            #           f"sum {np.sum(source_profile_frames[index_source_prof])}")
            #     # import matplotlib.pyplot as plt
                # plt.imshow(source_profile_frames[index_source_prof])
                # plt.title(
                #     f"{index_source_prof} / {source_profile_frames.shape[0] - 1} from {cinac_recording.identifier}")
                # plt.show()

            if np.sum(source_profile_frames) == 0:
                print(f"{cinac_recording.identifier} np.sum(source_profile_frames) == 0 in "
                      f"generate_movies_from_metadata() in cinac_movie_patch.py")

            input_index = 1

            use_the_whole_frame = False
            if use_the_whole_frame:
                # doing augmentation if the function exists
                if augmentation_fct is not None:
                    source_profile_frames = augmentation_fct(source_profile_frames)
                # then we fit it the frame use by the network, padding the surrounding by zero if necessary
                profile_fit = np.zeros((len(frames), self.max_height, self.max_width))
                # we center the source profile
                y_coord = (profile_fit.shape[1] - source_profile_frames.shape[1]) // 2
                x_coord = (profile_fit.shape[2] - source_profile_frames.shape[2]) // 2
                profile_fit[:, y_coord:source_profile_frames.shape[1] + y_coord,
                x_coord:source_profile_frames.shape[2] + x_coord] = \
                    source_profile_frames

                profile_fit = profile_fit.reshape((profile_fit.shape[0], profile_fit.shape[1], profile_fit.shape[2], 1))
                data = inputs_dict[f"input_{input_index}"]
                data[index_batch] = profile_fit
                input_index += 1

            # then we compute the frame with just the mask of each cell (the main one (with input_0 index) and the ones
            # that overlaps)
            mask_source_profiles_keys = np.array(list(mask_source_profiles.keys()))

            mask_for_all_cells = np.zeros((source_profile_frames.shape[1], source_profile_frames.shape[2]),
                                          dtype="int8")
            if self.with_neuropil_mask:
                neuropil_mask = np.zeros((source_profile_frames.shape[1], source_profile_frames.shape[2]),
                                         dtype="int8")
            for cell_index, mask_source_profile in mask_source_profiles.items():
                if cell_index == cell:
                    source_profile_frames_masked = np.copy(source_profile_frames)
                    source_profile_frames_masked[:, mask_source_profile] = 0
                    if self.with_neuropil_mask:
                        # print(f"source_profile_frames.shape {source_profile_frames.shape}, "
                        #       f"mask_source_profile.shape {mask_source_profile.shape}, "
                        #       f"neuropil_mask.shape {(1 - mask_source_profile).shape}")
                        neuropil_mask[1 - mask_source_profile] = 1

                    # doing augmentation if the function exists
                    if augmentation_fct is not None:
                        source_profile_frames_masked = augmentation_fct(source_profile_frames_masked)

                    # then we fit it the frame use by the network, padding the surrounding by zero if necessary
                    profile_fit_masked = np.zeros((len(frames), self.max_height, self.max_width))
                    # we center the source profile
                    y_coord = (profile_fit_masked.shape[1] - source_profile_frames.shape[1]) // 2
                    x_coord = (profile_fit_masked.shape[2] - source_profile_frames.shape[2]) // 2
                    profile_fit_masked[:, y_coord:source_profile_frames.shape[1] + y_coord,
                    x_coord:source_profile_frames.shape[2] + x_coord] = \
                        source_profile_frames_masked

                    profile_fit_masked = profile_fit_masked.reshape((profile_fit_masked.shape[0],
                                                                     profile_fit_masked.shape[1],
                                                                     profile_fit_masked.shape[2], 1))

                    inputs_dict["input_0"][index_batch] = profile_fit_masked

                    continue
                else:
                    # mask_source_profile worth zero for the pixels in the cell
                    mask_for_all_cells[1 - mask_source_profile] = 1
                    if self.with_neuropil_mask:
                        neuropil_mask[1 - mask_source_profile] = 1

            # now feeding it with the overlaping cells mask
            if len(mask_source_profiles) > 0:
                source_profile_frames_masked = np.copy(source_profile_frames)
                source_profile_frames_masked[:, 1 - mask_for_all_cells] = 0

                # doing augmentation if the function exists
                if augmentation_fct is not None:
                    source_profile_frames_masked = augmentation_fct(source_profile_frames_masked)

                # then we fit it the frame use by the network, padding the surrounding by zero if necessary
                profile_fit_masked = np.zeros((len(frames), self.max_height, self.max_width))
                # we center the source profile
                y_coord = (profile_fit_masked.shape[1] - source_profile_frames.shape[1]) // 2
                x_coord = (profile_fit_masked.shape[2] - source_profile_frames.shape[2]) // 2
                profile_fit_masked[:, y_coord:source_profile_frames.shape[1] + y_coord,
                x_coord:source_profile_frames.shape[2] + x_coord] = \
                    source_profile_frames_masked

                profile_fit_masked = profile_fit_masked.reshape((profile_fit_masked.shape[0],
                                                                 profile_fit_masked.shape[1],
                                                                 profile_fit_masked.shape[2], 1))

                inputs_dict["input_1"][index_batch] = profile_fit_masked
            else:
                # empty frame if there is not overlaping cell
                profile_fit_masked = np.zeros((len(frames), self.max_height, self.max_width))
                profile_fit_masked = profile_fit_masked.reshape((profile_fit_masked.shape[0],
                                                                 profile_fit_masked.shape[1],
                                                                 profile_fit_masked.shape[2], 1))

                inputs_dict["input_1"][index_batch] = profile_fit_masked

            # now feeding it with the neuropil mask
            if self.with_neuropil_mask:
                source_profile_frames_masked = np.copy(source_profile_frames)
                # "deleting" the cells
                source_profile_frames_masked[:, neuropil_mask] = 0

                # doing augmentation if the function exists
                if augmentation_fct is not None:
                    source_profile_frames_masked = augmentation_fct(source_profile_frames_masked)

                # then we fit it the frame use by the network, padding the surrounding by zero if necessary
                profile_fit_masked = np.zeros((len(frames), self.max_height, self.max_width))
                # we center the source profile
                y_coord = (profile_fit_masked.shape[1] - source_profile_frames.shape[1]) // 2
                x_coord = (profile_fit_masked.shape[2] - source_profile_frames.shape[2]) // 2
                profile_fit_masked[:, y_coord:source_profile_frames.shape[1] + y_coord,
                x_coord:source_profile_frames.shape[2] + x_coord] = \
                    source_profile_frames_masked

                profile_fit_masked = profile_fit_masked.reshape((profile_fit_masked.shape[0],
                                                                 profile_fit_masked.shape[1],
                                                                 profile_fit_masked.shape[2], 1))

                inputs_dict["input_2"][index_batch] = profile_fit_masked

        # print(f"End generate_movies_from_metadata() in cinac_movie_patch.py")

        if with_labels:
            return inputs_dict, labels
        else:
            return inputs_dict

    def __str__(self):
        bonus_str = ""
        if self.with_neuropil_mask:
            bonus_str = " + one with neuropil mask"
        return f"{self.n_inputs} inputs. Main cell mask + one with all overlaping cells mask{bonus_str}"


class MoviePatchData:

    def __init__(self, cinac_recording, cell, index_movie, max_n_transformations,
                 encoded_frames, decoding_frame_dict, window_len, cell_type_classifier_mode=False,
                 session_id=None, with_info=False, to_keep_absolutely=False,
                 ground_truth=None):
        """

        Args:
            cinac_recording:
            cell:
            index_movie:
            max_n_transformations:
            encoded_frames: could be None if with_info is False
            decoding_frame_dict:could be None if with_info is False
            window_len:
            cell_type_classifier_mode: boolean, indicate if we should classify cell type or cell activity
            session_id: a string representing the id of the session. If None, it will be given a default.
            Session_id is used to stratify movie_patch according to their session, so sessions are balanced between
            themselves. Useful for training data, not to predict
            with_info: to know how many transients in this frame
            to_keep_absolutely:
            ground_truth: if cell activity classifier represents a 1d array, same length as window_len,
            give for each frame of the given cell, if it is active (1) or non active (0). If we use it as a cell type
            classifier then the value is encoded as 0 or 1 for a single class or more integers values for a multi-class
            classifier
        """
        # max_n_transformationsmax number of transformations to a movie patch
        # if the number of available function to transform is lower, the lower one would be kept
        self.manual_max_transformation = max_n_transformations
        self.cinac_recording = cinac_recording
        self.recording_identifier = cinac_recording.identifier
        self.cell = cell
        # session id, allows to stratify data according to the session of the recording
        self.session_id = session_id if session_id is not None else "jon_doe"
        # index of the first frame of the movie over the whole movie
        self.index_movie = index_movie
        self.last_index_movie = index_movie + window_len - 1
        self.window_len = window_len
        self.ground_truth = ground_truth
        self.cell_type_classifier_mode = cell_type_classifier_mode
        if self.ground_truth is not None and (not self.cell_type_classifier_mode):
            if len(self.ground_truth) != self.window_len:
                raise Exception("Ground_truth and window_len should be the same length. "
                                 "Here {len(self.ground_truth)} != {len(self.window_len)}")
        # weight to apply, use by the model to produce the loss function result
        self.weight = 1
        # means it's an import movie patch and that it should not be deleted during stratification
        # also it would have a minimum number of transformation
        self.to_keep_absolutely = to_keep_absolutely
        # number of transformation to perform on this movie, information to use if with_info == True
        # otherwise it means the object will be transform with the self.data_augmentation_fct
        if self.to_keep_absolutely:
            self.n_augmentations_to_perform = 3
        else:
            self.n_augmentations_to_perform = 0

        # used if a movie_data has been copied
        self.data_augmentation_fct = None

        # set of functions used for data augmentation, one will be selected when copying a movie
        self.data_augmentation_fct_list = list()
        # functions based on rotations and flips
        rot_fct = []
        # adding fct to the set
        flips = [horizontal_flip, vertical_flip, v_h_flip]
        for flip in flips:
            rot_fct.append(flip)
        # 180° angle is the same as same as v_h_flip
        # 10 angles
        rotation_angles = np.array([20, 50, 90, 120, 160, 200, 230, 270, 310, 240])
        np.random.shuffle(rotation_angles)
        for angle in rotation_angles:
            rot_fct.append(lambda movie: rotate_movie(movie, angle))
        # 24 shifting transformations combinaison
        x_shift_y_shift_couples = []
        for x_shift in np.arange(-2, 3):
            for y_shift in np.arange(-2, 3):
                if (x_shift == 0) and (y_shift == 0):
                    continue
                x_shift_y_shift_couples.append((x_shift, y_shift))
        shifts_fct = []
        # keeping 11 shifts, from random
        n_shifts = 11
        shift_indices = np.arange(len(x_shift_y_shift_couples))
        if n_shifts < len(shift_indices):
            np.random.shuffle(shift_indices)
            shift_indices = shift_indices[:n_shifts]
        for index in shift_indices:
            x_shift = x_shift_y_shift_couples[index][0]
            y_shift = x_shift_y_shift_couples[index][1]
            shifts_fct.append(lambda movie: shift_movie(movie, x_shift=x_shift, y_shift=x_shift))

        for i in np.arange(max(len(rot_fct), len(shifts_fct))):
            if i < len(rot_fct):
                self.data_augmentation_fct_list.append(rot_fct[i])
            if i < len(shifts_fct):
                self.data_augmentation_fct_list.append(shifts_fct[i])

        self.n_available_augmentation_fct = min(self.manual_max_transformation, len(self.data_augmentation_fct_list))

        # movie_info dict containing the different informations about the movie such as the number of transients etc...
        """
        Keys so far for self.movie_info (with value type) -> comments :

        n_transient (int)
        transients_lengths (list of int)
        transients_amplitudes (list of float)
        n_cropped_transient (int) -> max value should be 2
        cropped_transients_lengths (list of int)
        n_fake_transient (int)
        n_cropped_fake_transient (int) > max value should be 2
        fake_transients_lengths (list of int)
        fake_transients_amplitudes (list of float)
        """
        self.movie_info = None
        self.encoded_frames = encoded_frames
        self.decoding_frame_dict = decoding_frame_dict
        if with_info:
            self.movie_info = dict()
            # then we want to know how many transients in this frame etc...
            # each code represent a specific event
            unique_codes = np.unique(encoded_frames[index_movie:index_movie + window_len])
            # print(f"unique_codes {unique_codes},  len {len(unique_codes)}")
            is_only_neuropil = True
            for code in unique_codes:
                event = decoding_frame_dict[code]
                if not event.neuropil:
                    is_only_neuropil = False

                if event.real_transient or event.fake_transient:

                    # we need to determine if it's a cropped one or full one
                    if (event.first_frame_event < index_movie) or (event.last_frame_event > self.last_index_movie):
                        # it's cropped
                        if event.real_transient:
                            key_str = "n_cropped_transient"
                            if "cropped_transients_lengths" not in self.movie_info:
                                self.movie_info["cropped_transients_lengths"] = []
                            self.movie_info["cropped_transients_lengths"].append(event.length_event)
                            if "transients_amplitudes" not in self.movie_info:
                                self.movie_info["transients_amplitudes"] = []
                            self.movie_info["transients_amplitudes"].append(event.amplitude)
                        else:
                            key_str = "n_cropped_fake_transient"
                            if "fake_transients_amplitudes" not in self.movie_info:
                                self.movie_info["fake_transients_amplitudes"] = []
                            self.movie_info["fake_transients_amplitudes"].append(event.amplitude)
                        self.movie_info[key_str] = self.movie_info.get(key_str, 0) + 1
                        continue

                    # means it's a full transient
                    if event.real_transient:
                        key_str = "n_transient"
                        if "transients_lengths" not in self.movie_info:
                            self.movie_info["transients_lengths"] = []
                        self.movie_info["transients_lengths"].append(event.length_event)
                        if "transients_amplitudes" not in self.movie_info:
                            self.movie_info["transients_amplitudes"] = []
                        self.movie_info["transients_amplitudes"].append(event.amplitude)
                    else:
                        key_str = "n_fake_transient"
                        if "fake_transients_lengths" not in self.movie_info:
                            self.movie_info["fake_transients_lengths"] = []
                        self.movie_info["fake_transients_lengths"].append(event.length_event)
                        if "fake_transients_amplitudes" not in self.movie_info:
                            self.movie_info["fake_transients_amplitudes"] = []
                        self.movie_info["fake_transients_amplitudes"].append(event.amplitude)
                    self.movie_info[key_str] = self.movie_info.get(key_str, 0) + 1
            if is_only_neuropil:
                self.movie_info["only_neuropil"] = True

    def get_labels(self, using_multi_class):
        """
        Return the labels for this data, could be if the cell is active for any given frame or the cell
        type depending on the classifier mode
        Args:
            using_multi_class:

        Returns:

        """
        frames = np.arange(self.index_movie, self.last_index_movie + 1)
        if using_multi_class <= 1:
            return self.ground_truth
        else:
            if self.cell_type_classifier_mode:
                labels = np.zeros(using_multi_class, dtype="uint8")
                labels[self.ground_truth] = 1
                return labels

            if using_multi_class == 3:
                unique_codes = np.unique(self.encoded_frames[frames])
                labels = np.zeros((self.window_len, using_multi_class), dtype="uint8")
                # class 0: real transient
                # class 1: fake transient
                # class 2 is "unclassifierd" or "noise" that includes decay and neuropil
                for code in unique_codes:
                    movie_event = self.decoding_frame_dict[code]
                    if movie_event.real_transient:
                        labels[self.encoded_frames[frames] == code, 0] = 1
                    elif movie_event.fake_transient:
                        labels[self.encoded_frames[frames] == code, 1] = 1
                    else:
                        labels[self.encoded_frames[frames] == code, 2] = 1
                return labels
            else:
                raise Exception(f"using_multi_class {using_multi_class} not implemented yet")

    def __eq__(self, other):
        if self.recording_identifier != other.recording_identifier:
            return False
        if self.cell != other.cell:
            return False
        if self.index_movie != other.index_movie:
            return False
        return True

    def copy(self):
        movie_copy = MoviePatchData(cinac_recording=self.cinac_recording, cell=self.cell,
                                    index_movie=self.index_movie, session_id=self.session_id,
                                    max_n_transformations=self.manual_max_transformation,
                                    encoded_frames=self.encoded_frames, decoding_frame_dict=self.decoding_frame_dict,
                                    window_len=self.window_len, ground_truth=self.ground_truth,
                                    cell_type_classifier_mode=self.cell_type_classifier_mode,
                                    to_keep_absolutely=self.to_keep_absolutely)
        movie_copy.data_augmentation_fct = self.data_augmentation_fct
        return movie_copy

    def add_n_augmentation(self, n_augmentation):
        self.n_augmentations_to_perform = min(self.n_augmentations_to_perform + n_augmentation,
                                              self.n_available_augmentation_fct)

    def pick_a_transformation_fct(self):
        if len(self.data_augmentation_fct_list) > 0:
            fct = self.data_augmentation_fct_list[0]
            self.data_augmentation_fct_list = self.data_augmentation_fct_list[1:]
            return fct
        return None

    def is_only_neuropil(self):
        """

        :return: True if there is only neuropil (no transients), False otherwise
        """
        if self.movie_info is None:
            return False

        if "n_transient" in self.movie_info:
            return False
        if "n_cropped_transient" in self.movie_info:
            return False
        if "n_fake_transient" in self.movie_info:
            return False
        if "n_cropped_fake_transient" in self.movie_info:
            return False

        return True

# TODO: use tf.data instead of keras.utils.Sequence in the future
class DataGenerator(keras.utils.Sequence):
    """
    Based on an exemple found in https://stanford.edu/~shervine/blog/keras-how-to-generate-data-on-the-fly
    Feed to keras to generate data
    """

    # 'Generates data for Keras'
    def __init__(self, data_list, movie_patch_generator,
                 batch_size, window_len, with_augmentation,
                 pixels_around, buffer, max_width, max_height,
                 is_shuffle=True):
        """

        :param data_list: a list containing the information to get the data. Each element
        is an instance of MoviePatchData
        :param batch_size:
        :param window_len:
        :param with_augmentation:
        :param is_shuffle:
        :param max_width:
        :param max_height:
        :param cell_type_classifier_mode: indicate the labeling is regarding what is the cell type and not is the cell
        active
        """
        # 'Initialization'
        self.max_width = max_width
        self.max_height = max_height
        self.pixels_around = pixels_around
        self.buffer = buffer
        self.window_len = window_len
        self.input_shape = (self.window_len, self.max_height, self.max_width, 1)
        self.batch_size = batch_size
        self.data_list = data_list
        self.with_augmentation = with_augmentation
        self.movie_patch_generator = movie_patch_generator
        # to improve performance, keep in memory the mask_profile of a cell and the coords of the frame surrounding the
        # the cell, the key is a string ms.description + cell
        # not used anymore
        # self.source_profiles_dict = dict()

        if self.with_augmentation:
            # augment the dict now, adding to the key a str representing the transformation and same for
            # in the value
            self.prepare_augmentation()

        # useful for the shuffling
        self.n_samples = len(self.data_list)
        # self.n_channels = n_channels
        # self.n_classes = n_classes
        self.is_shuffle = is_shuffle
        self.indexes = None
        self.on_epoch_end()

    def prepare_augmentation(self):
        n_samples = len(self.data_list)
        print(f"n_samples before data augmentation: {n_samples}")
        new_data = []

        for index_data in np.arange(n_samples):
            movie_data = self.data_list[index_data]
            # we will do as many transformation as indicated in movie_data.n_augmentations_to_perform
            if movie_data.n_augmentations_to_perform == 0:
                continue
            for t in np.arange(movie_data.n_augmentations_to_perform):
                if t >= movie_data.n_available_augmentation_fct:
                    break
                new_movie = movie_data.copy()
                new_movie.data_augmentation_fct = movie_data.pick_a_transformation_fct()
                new_data.append(new_movie)

        self.data_list.extend(new_data)

        print(f"n_samples after data augmentation: {len(self.data_list)}")

    def __len__(self):
        # 'Denotes the number of batches per epoch'
        return int(np.floor(self.n_samples / self.batch_size))

    def __getitem__(self, index):
        # 'Generate one batch of data'
        # Generate indexes of the batch
        indexes = self.indexes[index * self.batch_size:(index + 1) * self.batch_size]
        # print(f"len(indexes) {len(indexes)}")
        # Find list of IDs
        data_list_tmp = [self.data_list[k] for k in indexes]

        # Generate data
        data, labels, sample_weights = self.__data_generation(data_list_tmp)

        return data, labels, sample_weights

    def on_epoch_end(self):
        # 'Updates indexes after each epoch'
        self.indexes = np.arange(self.n_samples)
        # TODO: each mini-batch should have the same proportion of data (neuropil, transients, fake transients)
        # TODO: create a function that shuffle the index with respect of this information
        if self.is_shuffle:
            np.random.shuffle(self.indexes)
        # self.data_keys = list(self.data_dict.keys())
        # if self.is_shuffle:
        #     shuffle(self.data_keys)

    def __data_generation(self, data_list_tmp):
        # len(data_list_tmp) == self.batch_size
        # 'Generates data containing batch_size samples' # data : (self.batch_size, *dim, n_channels)
        # Initialization

        # data, data_masked, labels = generate_movies_from_metadata(movie_data_list=data_list_tmp,
        #                                                           window_len=self.window_len,
        #                                                           max_width=self.max_width,
        #                                                           max_height=self.max_height,
        #                                                           pixels_around=self.pixels_around,
        #                                                           buffer=self.buffer,
        #                                                           source_profiles_dict=self.source_profiles_dict)
        data_dict, labels = self.movie_patch_generator.generate_movies_from_metadata(movie_data_list=data_list_tmp,
                                                                                     memory_dict=None)
        # print(f"__data_generation data.shape {data.shape}")
        # put more weight to the active frames
        # TODO: reshape labels such as shape is (batch_size, window_len, 1) and then use "temporal" mode in compile
        # TODO: otherwise, use the weight in the movie_data in data_list_tmp to apply the corresponding weight
        # sample_weights = np.ones(labels.shape)
        # sample_weights[labels == 1] = 5
        sample_weights = np.ones(labels.shape[0])

        for index_batch, movie_data in enumerate(data_list_tmp):
            sample_weights[index_batch] = movie_data.weight

        return data_dict, labels, sample_weights



import sys
if sys.version_info[0] < 3:
    from Tkinter import *
else:
    from tkinter import *

from deepcinac.gui.cinac_gui import ChoiceFormatFrame

root = Tk()
root.title(f"Options")

app = ChoiceFormatFrame(master=root, default_path=None)
app.mainloop()



import numpy as np
from ScanImageTiffReader import ScanImageTiffReader
import time
import os
import PIL
from PIL import ImageSequence, ImageDraw
import tifffile
from shapely import geometry
import scipy.signal
from scipy import ndimage


def smooth_convolve(x, window_len=11, window='hanning'):
    """smooth the data using a window with requested size.

    This method is based on the convolution of a scaled window with the signal.
    The signal is prepared by introducing reflected copies of the signal
    (with the window size) in both ends so that transient parts are minimized
    in the begining and end part of the output signal.

    input:
        x: the input signal
        window_len: the dimension of the smoothing window; should be an odd integer
        window: the type of window from 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'
            flat window will produce a moving average smoothing.

    output:
        the smoothed signal

    example:

    t=linspace(-2,2,0.1)
    x=sin(t)+randn(len(t))*0.1
    y=smooth(x)

    Source: https://scipy-cookbook.readthedocs.io/items/SignalSmooth.html

    NOTE: length(output) != length(input), to correct this: return y[(window_len/2-1):-(window_len/2)] instead of just y.
    """

    if x.ndim != 1:
        raise ValueError("smooth only accepts 1 dimension arrays.")

    if x.size < window_len:
        raise ValueError("Input vector needs to be bigger than window size.")

    if window_len < 3:
        return x

    if not window in ['flat', 'hanning', 'hamming', 'bartlett', 'blackman']:
        raise ValueError("Window is on of 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'")

    s = np.r_[x[window_len - 1:0:-1], x, x[-2:-window_len - 1:-1]]
    # print(len(s))
    if window == 'flat':  # moving average
        w = np.ones(window_len, 'd')
    else:
        w = eval('np.' + window + '(window_len)')

    y = np.convolve(w / w.sum(), s, mode='valid')

    return y


def get_continous_time_periods(binary_array):
    """
    take a binary array and return a list of tuples representing the first and last position(included) of continuous
    positive period
    This code was copied from another project or from a forum, but i've lost the reference.
    :param binary_array:
    :return:
    """
    binary_array = np.copy(binary_array).astype("int8")
    n_times = len(binary_array)
    d_times = np.diff(binary_array)
    # show the +1 and -1 edges
    pos = np.where(d_times == 1)[0] + 1
    neg = np.where(d_times == -1)[0] + 1

    if (pos.size == 0) and (neg.size == 0):
        if len(np.nonzero(binary_array)[0]) > 0:
            return [(0, n_times-1)]
        else:
            return []
    elif pos.size == 0:
        # i.e., starts on an spike, then stops
        return [(0, neg[0])]
    elif neg.size == 0:
        # starts, then ends on a spike.
        return [(pos[0], n_times-1)]
    else:
        if pos[0] > neg[0]:
            # we start with a spike
            pos = np.insert(pos, 0, 0)
        if neg[-1] < pos[-1]:
            #  we end with aspike
            neg = np.append(neg, n_times - 1)
        # NOTE: by this time, length(pos)==length(neg), necessarily
        h = np.matrix([pos, neg])
        if np.any(h):
            result = []
            for i in np.arange(h.shape[1]):
                if h[1, i] == n_times-1:
                    result.append((h[0, i], h[1, i]))
                else:
                    result.append((h[0, i], h[1, i]-1))
            return result
    return []


def find_all_onsets_and_peaks_on_fluorescence_signal(smooth_trace, threshold_factor=0.5, identifier=None):
    """
    Get all potential onsets and peaks from a fluorescence signal
    Args:
        smooth_trace: fluorescence signal of cell, should be smooth
        threshold_factor: use to define a threshold over which to keep peaks.
        The threshold used is (threshold_factor * std(smooth_trace) + min(smooth_trace)
        identifier (str) for debugging purpose

    Returns: a 1d array of integers (binary) representing the time  when the cell is active

    """

    n_frames = len(smooth_trace)
    peak_nums = np.zeros(n_frames, dtype="int8")
    peaks, properties = scipy.signal.find_peaks(x=smooth_trace, distance=2)
    peak_nums[peaks] = 1
    spike_nums = np.zeros(n_frames, dtype="int8")
    onsets = []
    diff_values = np.diff(smooth_trace)
    for index, value in enumerate(diff_values):
        if index == (len(diff_values) - 1):
            continue
        if value < 0:
            if diff_values[index + 1] >= 0:
                onsets.append(index + 1)
    # print(f"onsets {len(onsets)}")
    onsets = np.array(onsets)
    if len(onsets) == 0:
        print(f"find_all_onsets_and_peaks_on_fluorescence_signal() no onsets found over {n_frames} "
              f"frames in {identifier}, mean smooth traces {np.mean(smooth_trace)}")
        return np.zeros(n_frames, dtype="int8")
    spike_nums[onsets] = 1

    threshold = (threshold_factor * np.std(smooth_trace)) + np.min(smooth_trace)
    peaks_under_threshold_index = peaks[smooth_trace[peaks] < threshold]
    # peaks_over_threshold_index = peaks[trace[peaks] >= threshold]
    # removing peaks under threshold and associated onsets
    peak_nums[peaks_under_threshold_index] = 0

    # onsets to remove
    onsets_index = np.where(spike_nums)[0]
    onsets_detected = []
    for peak_time in peaks_under_threshold_index:
        # looking for the peak preceding the onset
        onsets_before = np.where(onsets_index < peak_time)[0]
        if len(onsets_before) > 0:
            onset_to_remove = onsets_index[onsets_before[-1]]
            onsets_detected.append(onset_to_remove)
    # print(f"onsets_detected {onsets_detected}")
    if len(onsets_detected) > 0:
        spike_nums[np.array(onsets_detected)] = 0

    # now we construct the spike_nums_dur
    spike_nums_dur = np.zeros(n_frames, dtype="int8")

    peaks_index = np.where(peak_nums)[0]
    onsets_index = np.where(spike_nums)[0]

    for onset_index in onsets_index:
        peaks_after = np.where(peaks_index > onset_index)[0]
        if len(peaks_after) == 0:
            continue
        peaks_after = peaks_index[peaks_after]
        peak_after = peaks_after[0]
        spike_nums_dur[onset_index:peak_after + 1] = 1

    return spike_nums_dur

def check_one_dir_by_id_exists(identifiers, results_path, dir_in_id_name=False):
    """
    Check if for each identifier in idenfifiers, a dir with this name exists
    in results_path. If they all exists, then True is return, otherwise False
    Useful to check if a CI_movie has been separated in multiple tiff files.
    However, doesn't check in all the files are present, we assume that they are
    Args:
        identifiers: list of string
        results_path: a path where to check for existing dir
        dir_in_id_name: if True, means the dir name should be in the identifier, no need for the directory to be
        exactly named as identifier

    Returns: boolean

    """
    # first we get a list of all movies already splited as tiff files
    dir_names = []
    # look for filenames in the fisst directory, if we don't break, it will go through all directories
    for (dirpath, dirnames, local_filenames) in os.walk(results_path):
        dir_names.extend([x.lower() for x in dirnames])
        break

    for identifier in identifiers:
        if dir_in_id_name:
            not_found = True
            for dir_name in dir_names:
                if dir_name in identifier.lower():
                    not_found = False
                    break
            if not_found:
                print(f"check_one_dir_by_id_exists(): identifier {identifier} NOT FOUND in {results_path}")
                return False
        else:
            if identifier.lower() not in dir_names:
                return False

    return True


def create_one_npy_file_by_frame(movies_to_split, results_path,
                                without_mean_std_files=False, verbose=0):
    """
    Split a calcium imaging movie so that each frame will be saved as a npy file.
    If the directory for results already contains a directory of the name of the movie identifier,
    the data won't be erased.
    Args:
        movies_to_split: a dictionary with as a key an identifier for the movie that will be used to name
        the directory in which the tiff will be put (using lower case), and as value either the file_name
        of the tiff or an numpy float
        ndarray representing the calcium imaging data (should n_frames * n_pixels_x * n_pixels_y)
        results_path: String. The directory in which will be created the directories containing the tiff files.
        verbose: Integer. 0, 1, or 2. Verbosity mode. 0 = silent, 1 = times for main operation, 2 = various prints.
        without_mean_std_files: if True, mean and std are not recorded, useful if the movie is already normalized
    Returns: None

    """

    if verbose == 2:
        print("## function create_one_npy_file_by_frame()")

    # first we get a list of all movies already splited as tiff files
    dir_names = []
    # look for filenames in the fisst directory, if we don't break, it will go through all directories
    for (dirpath, dirnames, local_filenames) in os.walk(results_path):
        dir_names.extend([x.lower() for x in dirnames])
        break

    for identifier, movie_value in movies_to_split.items():
        if identifier.lower() in dir_names:
            if verbose == 2:
                print(f"Numpy files for {identifier} already existing")
            # it means we've already created the tiffs
            continue

        if verbose == 2:
            print(f"Creating numpy files for {identifier}")
        if isinstance(movie_value, str):
            try:
                start_time = time.time()
                tiff_movie = ScanImageTiffReader(movie_value).data()
                n_frames = len(tiff_movie)
                stop_time = time.time()
                if verbose >= 1:
                    print(f"Time for loading movie with scan_image_tiff: "
                          f"{np.round(stop_time - start_time, 3)} s")
            except Exception as e:
                start_time = time.time()
                im = PIL.Image.open(movie_value)
                n_frames = len(list(ImageSequence.Iterator(im)))
                dim_y, dim_x = np.array(im).shape
                if verbose == 2:
                    print(f"n_frames {n_frames}, dim_x {dim_x}, dim_y {dim_y}")
                tiff_movie = np.zeros((n_frames, dim_y, dim_x), dtype="uint16")
                for frame, page in enumerate(ImageSequence.Iterator(im)):
                    tiff_movie[frame] = np.array(page)
                stop_time = time.time()
                if verbose >= 1:
                    print(f"Time for loading movie: "
                          f"{np.round(stop_time - start_time, 3)} s")
        else:
            tiff_movie = movie_value
            n_frames = len(tiff_movie)

        ms_path = os.path.join(results_path, identifier.lower())
        os.mkdir(ms_path)

        if not without_mean_std_files:
            # saving the mean and std value of the movie, for later on normalization.
            mean_value = np.mean(tiff_movie)
            std_value = np.std(tiff_movie)
            np.save(os.path.join(ms_path, "mean.npy"), mean_value)
            np.save(os.path.join(ms_path, "std.npy"), std_value)

        start_time = time.time()
        # then saving each frame as a unique numpy file
        for frame in np.arange(n_frames):
            npy_file_name = os.path.join(ms_path, f"{frame}.npy")
            np.save(npy_file_name, tiff_movie[frame])
        stop_time = time.time()
        if verbose >= 1:
            print(f"Time for writing the npy frames: "
                  f"{np.round(stop_time - start_time, 3)} s")

        # to free memory
        if isinstance(movie_value, str):
            del tiff_movie


def create_one_tiff_file_by_frame(movies_to_split, results_path,
                                  without_mean_std_files=False, verbose=0):
    """
    Split a calcium imaging movie so that each frame will be saved as a tiff image.
    If the directory for results already contains a directory of the name of the movie identifier,
    the data won't be erased.
    Args:
        movies_to_split: a dictionary with as a key an identifier for the movie that will be used to name
        the directory in which the tiff will be put (using lower case), and as value either the file_name
        of the tiff or an numpy float
        ndarray representing the calcium imaging data (should n_frames * n_pixels_x * n_pixels_y)
        results_path: String. The directory in which will be created the directories containing the tiff files.
        verbose: Integer. 0, 1, or 2. Verbosity mode. 0 = silent, 1 = times for main operation, 2 = various prints.
        without_mean_std_files: if True, mean and std are not recorded, useful if the movie is already normalized
    Returns: None

    """

    if verbose == 2:
        print("create_one_tiff_file_by_frame")

    # first we get a list of all movies already splited as tiff files
    dir_names = []
    # look for filenames in the fisst directory, if we don't break, it will go through all directories
    for (dirpath, dirnames, local_filenames) in os.walk(results_path):
        dir_names.extend([x.lower() for x in dirnames])
        break

    for identifier, movie_value in movies_to_split.items():
        if identifier.lower() in dir_names:
            if verbose == 2:
                print(f"Tiffs for {identifier} already existing")
            # it means we've already created the tiffs
            continue

        if verbose == 2:
            print(f"Creating tiffs for {identifier}")
        if isinstance(movie_value, str):
            try:
                start_time = time.time()
                tiff_movie = ScanImageTiffReader(movie_value).data()
                n_frames = len(tiff_movie)
                stop_time = time.time()
                if verbose >= 1:
                    print(f"Time for loading movie with scan_image_tiff: "
                          f"{np.round(stop_time - start_time, 3)} s")
            except Exception as e:
                start_time = time.time()
                im = PIL.Image.open(movie_value)
                n_frames = len(list(ImageSequence.Iterator(im)))
                dim_y, dim_x = np.array(im).shape
                if verbose == 2:
                    print(f"n_frames {n_frames}, dim_x {dim_x}, dim_y {dim_y}")
                tiff_movie = np.zeros((n_frames, dim_y, dim_x), dtype="uint16")
                for frame, page in enumerate(ImageSequence.Iterator(im)):
                    tiff_movie[frame] = np.array(page)
                stop_time = time.time()
                if verbose >= 1:
                    print(f"Time for loading movie: "
                          f"{np.round(stop_time - start_time, 3)} s")
        else:
            tiff_movie = movie_value
            n_frames = len(tiff_movie)

        ms_path = os.path.join(results_path, identifier.lower())
        os.mkdir(ms_path)

        if not without_mean_std_files:
            # saving the mean and std value of the movie, for later on normalization.
            mean_value = np.mean(tiff_movie)
            std_value = np.std(tiff_movie)
            np.save(os.path.join(ms_path, "mean.npy"), mean_value)
            np.save(os.path.join(ms_path, "std.npy"), std_value)

        start_time = time.time()
        # then saving each frame as a unique tiff
        for frame in np.arange(n_frames):
            tiff_file_name = os.path.join(ms_path, f"{frame}.tiff")
            with tifffile.TiffWriter(tiff_file_name) as tiff:
                tiff.save(tiff_movie[frame], compress=0)
        stop_time = time.time()
        if verbose >= 1:
            print(f"Time for writing the tiffs: "
                  f"{np.round(stop_time - start_time, 3)} s")

        # to free memory
        if isinstance(movie_value, str):
            del tiff_movie


def scale_polygon_to_source(polygon, minx, miny):
    """
    Take an instance of shapely.geometry.Polygon or shapely.geometry.LineString and scale it, so that each of
     it coordinates are substracted by minx and miny on the x and y axis respectively
    coordinates to match minx and miny
    Args:
        polygon: Polygon instance from shapely package. Could also be a LineString
        minx: integer
        miny: integer

    Returns: a new shapely.geometry.Polygon or shapely.geometry.LineString

    """
    if isinstance(polygon, geometry.LineString):
        coords = list(polygon.coords)
    else:
        coords = list(polygon.exterior.coords)

    scaled_coords = []
    for coord in coords:
        scaled_coords.append((coord[0] - minx, coord[1] - miny))

    if isinstance(polygon, geometry.LineString):
        return geometry.LineString(scaled_coords)
    else:
        return geometry.Polygon(scaled_coords)


def load_movie(file_name, with_normalization, both_instances=False, verbose=True):
    """
    Load a movie in memory. So far only tiff format is valid
    Args:
        file_name:
        with_normalization: if True, normalize the movie using z-score formula
        both_instances: if with_normalization is True and both_instances is True, then
        the function return both movie, the normal and the normalized one in that order

    Returns:

    """
    try:
        start_time = time.time()
        tiff_movie = ScanImageTiffReader(file_name).data()
        stop_time = time.time()
        if verbose:
            print(f"Time for loading movie with ScanImageTiffReader: "
                  f"{np.round(stop_time - start_time, 3)} s")
    except Exception as e:
        start_time = time.time()
        im = PIL.Image.open(file_name)
        n_frames = len(list(ImageSequence.Iterator(im)))
        dim_y, dim_x = np.array(im).shape
        tiff_movie = np.zeros((n_frames, dim_y, dim_x), dtype="uint16")
        for frame, page in enumerate(ImageSequence.Iterator(im)):
            tiff_movie[frame] = np.array(page)
        stop_time = time.time()
        if verbose:
            print(f"Time for loading movie with PIL: "
                  f"{np.round(stop_time - start_time, 3)} s")
    if with_normalization:
        tiff_movie_normalized = tiff_movie - np.mean(tiff_movie)
        tiff_movie_normalized = tiff_movie_normalized / np.std(tiff_movie)
        if both_instances:
            return tiff_movie, tiff_movie_normalized
        return tiff_movie_normalized
    else:
        return tiff_movie


def build_raster_dur_from_onsets_peaks(onsets, peaks):
    """
    Build a raster_dur, a 2d binary array indicating when the cell is active (rise time).
    n_cells * n_frames
    Args:
        onsets: 2d binary array, n_cells * n_frames, 1 if onset at this frame
        peaks: 2d binary array, n_cells * n_frames

    Returns:

    """
    n_cells = len(onsets)
    n_frames = onsets.shape[1]
    raster_dur = np.zeros((n_cells, n_frames), dtype="int8")
    for cell in np.arange(n_cells):
        peaks_index = np.where(peaks[cell, :])[0]
        onsets_index = np.where(onsets[cell, :])[0]

        for onset_index in onsets_index:
            peaks_after = np.where(peaks_index > onset_index)[0]
            if len(peaks_after) == 0:
                continue
            peaks_after = peaks_index[peaks_after]
            peak_after = peaks_after[0]

            raster_dur[cell, onset_index:peak_after + 1] = 1
    return raster_dur

def get_source_profile_param(cell, movie_dimensions, coord_obj, max_width, max_height, pixels_around=0,
                             buffer=None, with_all_masks=False, get_only_polygon_contour=False):
    """
     For given cell, get the binary mask representing this cell with the possibility to get the binary masks
     of the cells it intersects with.

    Args:
        cell:
        movie_dimensions: tuple of integers, width and height of the movie
        coord_obj: instance of
        max_width: Max width of the frame returned. Might cropped some overlaping cell if necessary
        max_height: Max height of the frame returned. Might cropped some overlaping cell if necessary
        pixels_around: how many pixels to add around the frame containing the cell and the overlapping one,
        doesn't change the mask
        buffer: How much pixels to scale the cell contour in the mask. If buffer is 0 or None, then size of the cell
        won't change.
        with_all_masks: Return a dict with all overlaps cells masks + the main cell mask. The key is an int.
     The mask consist on a binary array of with 0 for all pixels in the cell, 1 otherwise
        get_only_polygon_contour: the mask represents then only the pixels that makes the contour of the cells

    Returns: A mask (numpy 2d binary array), with 0 for all pixels in the cell, 1 otherwise.
     A tuple with four integers representing the corner coordinates (minx, maxx, miny, maxy)

    """

    len_frame_x = movie_dimensions[1]
    len_frame_y = movie_dimensions[0]

    # determining the size of the square surrounding the cell so it includes all overlapping cells around
    overlapping_cells = coord_obj.intersect_cells.get(cell, [])
    cells_to_display = [cell]
    cells_to_display.extend(overlapping_cells)

    # calculating the bound that will surround all the cells
    # if the size of the frame already matches the max dimensions, then we keep it that way
    if (len_frame_y == max_width) and (len_frame_x == max_height):
        minx = 0
        miny = 0
        maxx = max_height - 1
        maxy = max_width - 1
        len_x = max_height
        len_y = max_width
    else:
        use_old_version = True
        if use_old_version:
            # calculating the bound that will surround all the cells
            minx = None
            maxx = None
            miny = None
            maxy = None

            centroid_cell = coord_obj.center_coord[cell]
            min_x_centroid = int(centroid_cell[0]) - int(max_height // 2)
            max_x_centroid = min_x_centroid + max_height - 1
            min_y_centroid = int(centroid_cell[1]) - int(max_width // 2)
            max_y_centroid = min_y_centroid + max_width - 1

            poly_gon_cell = coord_obj.cells_polygon[cell]
            minx_cell, miny_cell, maxx_cell, maxy_cell = np.array(list(poly_gon_cell.bounds)).astype(int)

            # if the cell is bigger than the square, then we just put its centroid in the middle of it
            if ((maxx_cell - minx_cell + 1) >= max_height) or \
                    ((maxy_cell - miny_cell + 1) >= max_width):
                minx = min_x_centroid
                miny = min_y_centroid

                # then we make sure we don't over go the border
                minx = max(0, minx)
                miny = max(0, miny)

                # then we make sure we don't over go the border
                if minx + max_height >= len_frame_x:
                    minx = len_frame_x - max_height - 1
                if miny + max_width >= len_frame_y:
                    miny = len_frame_y - max_width - 1

                maxx = minx + max_height - 1
                maxy = miny + max_width - 1

                len_x = max_height
                len_y = max_width
            else:

                for cell_in in cells_to_display:
                    poly_gon = coord_obj.cells_polygon[cell_in]

                    if minx is None:
                        minx, miny, maxx, maxy = np.array(list(poly_gon.bounds)).astype(int)
                    else:
                        tmp_minx, tmp_miny, tmp_maxx, tmp_maxy = np.array(list(poly_gon.bounds)).astype(int)
                        minx = min(minx, tmp_minx)
                        miny = min(miny, tmp_miny)
                        maxx = max(maxx, tmp_maxx)
                        maxy = max(maxy, tmp_maxy)

                # # centering the window on the cell
                # if (min_x_centroid <= minx) and \
                #         ((min_x_centroid + max_height) >= maxx):
                #     minx = min_x_centroid
                #
                # if (min_y_centroid <= miny) and \
                #         ((min_y_centroid + max_width) >= maxy):
                #     miny = min_y_centroid

                # trying to put the cell in, first by centering the frame around the centroid if necessary
                # then if some part of the cell still are out, we use the min of the cell
                if min_x_centroid < minx:
                    minx = min_x_centroid
                if minx_cell < minx:
                    minx = minx_cell
                if min_y_centroid < miny:
                    miny = min_y_centroid
                if miny_cell < miny:
                    miny = miny_cell

                if (minx + max_height - 1) < max_x_centroid:
                    minx = max_x_centroid - max_height + 1
                if (minx + max_height - 1) < maxx_cell:
                    minx = maxx_cell - max_height + 1
                if (maxy + max_width - 1) < max_y_centroid:
                    miny = max_y_centroid - max_width + 1
                if (miny + max_width - 1) < maxy_cell:
                    miny = maxy_cell - max_width + 1

                # then we make sure we don't over go the border
                minx = max(0, minx)
                miny = max(0, miny)
                if minx + max_height >= len_frame_x:
                    minx = len_frame_x - max_height - 1
                if miny + max_width >= len_frame_y:
                    miny = len_frame_y - max_width - 1

                maxx = minx + max_height - 1
                maxy = miny + max_width - 1
                len_x = max_height
                len_y = max_width
        else:
            centroid_cell = coord_obj.center_coord[cell]
            min_x_centroid = int(centroid_cell[0]) - int(max_height // 2)
            min_y_centroid = int(centroid_cell[1]) - int(max_width // 2)
            # we keep it simple and center the window around the centroid of the cell
            minx = min_x_centroid
            miny = min_y_centroid

            # then we make sure we don't go over the border
            minx = max(0, minx)
            miny = max(0, miny)
            if minx + max_height >= len_frame_x:
                minx = len_frame_x - max_height - 1
            if miny + max_width >= len_frame_y:
                miny = len_frame_y - max_width - 1

            maxx = minx + max_height - 1
            maxy = miny + max_width - 1
            len_x = max_height
            len_y = max_width

    mask_dict = dict()

    for cell_to_display in cells_to_display:
        if (not with_all_masks) and (cell_to_display != cell):
            continue
        polygon = coord_obj.cells_polygon[cell_to_display]
        # mask used in order to keep only the cells pixel
        # the mask put all pixels in the polygon, including the pixels on the exterior line to zero
        if minx == 0 and miny == 0:
            scaled_polygon = polygon
        else:
            scaled_polygon = scale_polygon_to_source(polygon=polygon, minx=minx, miny=miny)
        img = PIL.Image.new('1', (len_x, len_y), 1)
        if buffer is not None:
            scaled_polygon = scaled_polygon.buffer(buffer)
        fill_value = 0
        if get_only_polygon_contour:
            fill_value = None
        if isinstance(scaled_polygon, geometry.LineString):
            ImageDraw.Draw(img).polygon(list(scaled_polygon.coords), outline=0,
                                        fill=fill_value)
        else:
            ImageDraw.Draw(img).polygon(list(scaled_polygon.exterior.coords), outline=0,
                                        fill=fill_value)

        mask_dict[cell_to_display] = np.array(img)
        # print(f"(minx, maxx, miny, maxy) {(minx, maxx, miny, maxy)}")
        # print(f"im.shape {mask_dict[cell_to_display].shape}")

    if with_all_masks:
        return mask_dict, (minx, maxx, miny, maxy)
    else:
        return mask_dict[cell], (minx, maxx, miny, maxy)


def welsh_powell(graph):
    """
        implementation of welsh_powell algorithm
        https://github.com/MUSoC/Visualization-of-popular-algorithms-in-Python/blob/master/Graph%20Coloring/graph_coloring.py
        Args:
            graph: instance of networkx graph

        Returns:

    """
    # sorting the nodes based on it's valency
    node_list = sorted(graph.nodes(), key=lambda x: graph.degree(x))
    # dictionary to store the colors assigned to each node
    col_val = {}
    # assign the first color to the first node
    col_val[node_list[0]] = 0
    # Assign colors to remaining N-1 nodes
    for node in node_list[1:]:
        available = [True] * len(graph.nodes())  # boolean list[i] contains false if the node color 'i' is not available

        # iterates through all the adjacent nodes and marks it's color as unavailable, if it's color has been set already
        for adj_node in graph.neighbors(node):
            if adj_node in col_val.keys():
                col = col_val[adj_node]
                available[col] = False
        clr = 0
        for clr in range(len(available)):
            if available[clr] == True:
                break
        col_val[node] = clr

    return col_val

############################################################################################
############################################################################################
############################### data augmentation functions ###############################
############################################################################################
############################################################################################

def horizontal_flip(movie):
    """
    movie is a 3D numpy array
    :param movie:
    :return:
    """
    new_movie = np.zeros(movie.shape)
    for frame in np.arange(len(movie)):
        new_movie[frame] = np.fliplr(movie[frame])

    return new_movie


def vertical_flip(movie):
    """
    movie is a 3D numpy array
    :param movie:
    :return:
    """
    new_movie = np.zeros(movie.shape)
    for frame in np.arange(len(movie)):
        new_movie[frame] = np.flipud(movie[frame])

    return new_movie


def v_h_flip(movie):
    """
    movie is a 3D numpy array
    :param movie:
    :return:
    """
    new_movie = np.zeros(movie.shape)
    for frame in np.arange(len(movie)):
        new_movie[frame] = np.fliplr(np.flipud(movie[frame]))

    return new_movie


def rotate_movie(movie, angle):
    """
        movie is a 3D numpy array
        :param movie:
        :return:
        """
    new_movie = np.zeros(movie.shape)
    for frame in np.arange(len(movie)):
        new_movie[frame] = ndimage.rotate(movie[frame], angle=angle, reshape=False, mode='reflect')

    return new_movie


def shift_movie(movie, x_shift, y_shift):
    """
    movie is a 3D numpy array
    :param movie:
    :param x_shift:
    :param y_shift:
    :return:
    """
    if x_shift >= movie.shape[2]:
        raise Exception(f"x_shift {x_shift} >= movie.shape[2] {movie.shape[2]}")
    if y_shift >= movie.shape[1]:
        raise Exception(f"y_shift {y_shift} >= movie.shape[1] {movie.shape[1]}")

    new_movie = np.zeros(movie.shape)

    if (y_shift == 0) and (x_shift == 0):
        new_movie = movie[:, :, :]
    elif (y_shift == 0) and (x_shift > 0):
        for frame in np.arange(len(movie)):
            new_movie[frame, :, :-x_shift] = movie[frame, :, x_shift:]
    elif (y_shift == 0) and (x_shift < 0):
        for frame in np.arange(len(movie)):
            new_movie[frame, :, -x_shift:] = movie[frame, :, :x_shift]
    elif (y_shift > 0) and (x_shift == 0):
        for frame in np.arange(len(movie)):
            new_movie[frame, :-y_shift, :] = movie[frame, y_shift:, :]
    elif (y_shift < 0) and (x_shift == 0):
        for frame in np.arange(len(movie)):
            new_movie[frame, -y_shift:, :] = movie[frame, :y_shift, :]
    elif (y_shift > 0) and (x_shift > 0):
        for frame in np.arange(len(movie)):
            new_movie[frame, :-y_shift, :-x_shift] = movie[frame, y_shift:, x_shift:]
    elif (y_shift < 0) and (x_shift < 0):
        for frame in np.arange(len(movie)):
            new_movie[frame, -y_shift:, -x_shift:] = movie[frame, :y_shift, :x_shift]
    elif (y_shift > 0) and (x_shift < 0):
        for frame in np.arange(len(movie)):
            new_movie[frame, :-y_shift, -x_shift:] = movie[frame, y_shift:, :x_shift]
    elif (y_shift < 0) and (x_shift > 0):
        for frame in np.arange(len(movie)):
            new_movie[frame, -y_shift:, :-x_shift] = movie[frame, :y_shift, x_shift:]

    return new_movie



import numpy as np

def smooth_convolve(x, window_len=11, window='hanning'):
    """smooth the data using a window with requested size.

    This method is based on the convolution of a scaled window with the signal.
    The signal is prepared by introducing reflected copies of the signal
    (with the window size) in both ends so that transient parts are minimized
    in the begining and end part of the output signal.

    input:
        x: the input signal
        window_len: the dimension of the smoothing window; should be an odd integer
        window: the type of window from 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'
            flat window will produce a moving average smoothing.

    output:
        the smoothed signal

    example:

    t=linspace(-2,2,0.1)
    x=sin(t)+randn(len(t))*0.1
    y=smooth(x)

    see also:

    numpy.hanning, numpy.hamming, numpy.bartlett, numpy.blackman, numpy.convolve
    scipy.signal.lfilter

    TODO: the window parameter could be the window itself if an array instead of a string
    NOTE: length(output) != length(input), to correct this: return y[(window_len/2-1):-(window_len/2)] instead of just y.
    """

    if x.ndim != 1:
        raise ValueError("smooth only accepts 1 dimension arrays.")

    if x.size < window_len:
        raise ValueError("Input vector needs to be bigger than window size.")

    if window_len < 3:
        return x

    if not window in ['flat', 'hanning', 'hamming', 'bartlett', 'blackman']:
        raise ValueError("Window is on of 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'")

    s = np.r_[x[window_len - 1:0:-1], x, x[-2:-window_len - 1:-1]]
    # print(len(s))
    if window == 'flat':  # moving average
        w = np.ones(window_len, 'd')
    else:
        w = eval('np.' + window + '(window_len)')

    y = np.convolve(w / w.sum(), s, mode='valid')

    return y
